from functools import partial
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from bluesky.simulators import RunEngineSimulator
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.backlight import Backlight
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import PandAFastGridScan, ZebraFastGridScan
from dodal.devices.flux import Flux
from dodal.devices.i03.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.zocalo import ZocaloResults
from event_model import Event
from ophyd.sim import NullStatus
from ophyd_async.core import AsyncStatus
from ophyd_async.fastcs.panda import HDFPanda
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.xrc_result import XRayCentreResult
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    RobotLoadAndEnergyChangeComposite,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan import (
    RobotLoadThenCentreComposite,
)
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_gridscan_callbacks,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

FLYSCAN_RESULT_HIGH = XRayCentreResult(
    centre_of_mass_mm=np.array([0.1, 0.2, 0.3]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=30,
    total_count=100,
)
FLYSCAN_RESULT_MED = XRayCentreResult(
    centre_of_mass_mm=np.array([0.4, 0.5, 0.6]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=20,
    total_count=120,
)
FLYSCAN_RESULT_LOW = XRayCentreResult(
    centre_of_mass_mm=np.array([0.7, 0.8, 0.9]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=10,
    total_count=140,
)


def make_event_doc(data, descriptor="abc123") -> Event:
    return {
        "time": 0,
        "timestamps": {"a": 0},
        "seq_num": 0,
        "uid": "not so random uid",
        "descriptor": descriptor,
        "data": data,
    }


BASIC_PRE_SETUP_DOC = {
    "undulator-current_gap": 0,
    "synchrotron-synchrotron_mode": SynchrotronMode.USER,
    "s4_slit_gaps-xgap": 0,
    "s4_slit_gaps-ygap": 0,
    "smargon-x": 10.0,
    "smargon-y": 20.0,
    "smargon-z": 30.0,
}

BASIC_POST_SETUP_DOC = {
    "aperture_scatterguard-selected_aperture": ApertureValue.OUT_OF_BEAM,
    "aperture_scatterguard-radius": None,
    "aperture_scatterguard-aperture-x": 15,
    "aperture_scatterguard-aperture-y": 16,
    "aperture_scatterguard-aperture-z": 2,
    "aperture_scatterguard-scatterguard-x": 18,
    "aperture_scatterguard-scatterguard-y": 19,
    "attenuator-actual_transmission": 0,
    "flux-flux_reading": 10,
    "dcm-energy_in_kev": 11.105,
}


@pytest.fixture
def grid_detect_devices(
    aperture_scatterguard: ApertureScatterguard,
    backlight: Backlight,
    beamstop_i03: Beamstop,
    detector_motion: DetectorMotion,
    eiger: EigerDetector,
    smargon: Smargon,
    oav: OAV,
    ophyd_pin_tip_detection: PinTipDetection,
    zocalo: ZocaloResults,
    synchrotron: Synchrotron,
    fast_grid_scan: ZebraFastGridScan,
    s4_slit_gaps: S4SlitGaps,
    flux: Flux,
    zebra,
    zebra_shutter,
    xbpm_feedback,
    attenuator,
    undulator,
    undulator_dcm,
    dcm,
) -> GridDetectThenXRayCentreComposite:
    return GridDetectThenXRayCentreComposite(
        aperture_scatterguard=aperture_scatterguard,
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_i03,
        detector_motion=detector_motion,
        eiger=eiger,
        zebra_fast_grid_scan=fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=ophyd_pin_tip_detection,
        smargon=smargon,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        xbpm_feedback=xbpm_feedback,
        zebra=zebra,
        zocalo=zocalo,
        panda=MagicMock(spec=HDFPanda),
        panda_fast_grid_scan=MagicMock(spec=PandAFastGridScan),
        dcm=dcm,
        robot=MagicMock(spec=BartRobot),
        sample_shutter=zebra_shutter,
    )


@pytest.fixture
def sim_run_engine_for_rotation(sim_run_engine):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"values": {"value": SynchrotronMode.USER}},
        "synchrotron-synchrotron_mode",
    )
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"values": {"value": -1}},
        "synchrotron-top_up_start_countdown",
    )
    sim_run_engine.add_handler(
        "read", lambda msg: {"values": {"value": -1}}, "smargon_omega"
    )
    return sim_run_engine


def mock_zocalo_trigger(zocalo: ZocaloResults, result):
    @AsyncStatus.wrap
    async def mock_complete(results):
        await zocalo._put_results(results, {"dcid": 0, "dcgid": 0})

    zocalo.trigger = MagicMock(side_effect=partial(mock_complete, result))


def run_generic_ispyb_handler_setup(
    ispyb_handler: GridscanISPyBCallback,
    params: HyperionSpecifiedThreeDGridScan,
):
    """This is useful when testing 'run_gridscan_and_move(...)' because this stuff
    happens at the start of the outer plan."""

    ispyb_handler.active = True
    ispyb_handler.activity_gated_start(
        {
            "subplan_name": CONST.PLAN.GRIDSCAN_OUTER,
            "mx_bluesky_parameters": params.model_dump_json(),
        }  # type: ignore
    )
    ispyb_handler.activity_gated_descriptor(
        {"uid": "123abc", "name": CONST.DESCRIPTORS.HARDWARE_READ_PRE}  # type: ignore
    )
    ispyb_handler.activity_gated_event(
        make_event_doc(
            BASIC_PRE_SETUP_DOC,
            descriptor="123abc",
        )
    )
    ispyb_handler.activity_gated_descriptor(
        {"uid": "abc123", "name": CONST.DESCRIPTORS.HARDWARE_READ_DURING}  # type: ignore
    )
    ispyb_handler.activity_gated_event(
        make_event_doc(
            BASIC_POST_SETUP_DOC,
            descriptor="abc123",
        )
    )


def modified_store_grid_scan_mock(*args, dcids=(0, 0), dcgid=0, **kwargs):
    mock = MagicMock(spec=StoreInIspyb)
    mock.begin_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid
    )
    mock.update_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid, grid_ids=(0, 0)
    )
    return mock


@pytest.fixture
def mock_subscriptions(test_fgs_params):
    with (
        patch(
            "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
            autospec=True,
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.append_to_comment"
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.begin_deposition",
            new=MagicMock(
                return_value=IspybIds(
                    data_collection_ids=(0, 0), data_collection_group_id=0
                )
            ),
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.update_deposition",
            new=MagicMock(
                return_value=IspybIds(
                    data_collection_ids=(0, 0),
                    data_collection_group_id=0,
                    grid_ids=(0, 0),
                )
            ),
        ),
    ):
        nexus_callback, ispyb_callback = create_gridscan_callbacks()
        ispyb_callback.ispyb = MagicMock(spec=StoreInIspyb)

    return (nexus_callback, ispyb_callback)


def fake_read(obj, initial_positions, _):
    initial_positions[obj] = 0
    yield Msg("null", obj)


@pytest.fixture
def robot_load_composite(
    smargon,
    dcm,
    robot,
    aperture_scatterguard,
    oav,
    webcam,
    thawer,
    lower_gonio,
    eiger,
    xbpm_feedback,
    attenuator,
    beamstop_i03,
    fast_grid_scan,
    undulator,
    undulator_dcm,
    s4_slit_gaps,
    vfm,
    mirror_voltages,
    backlight,
    detector_motion,
    flux,
    pin_tip_detection_with_found_pin,
    zocalo,
    synchrotron,
    sample_shutter,
    zebra,
    panda,
    panda_fast_grid_scan,
) -> RobotLoadThenCentreComposite:
    set_mock_value(dcm.energy_in_kev.user_readback, 11.105)
    smargon.stub_offsets.set = MagicMock(return_value=NullStatus())
    aperture_scatterguard.set = MagicMock(return_value=NullStatus())
    set_mock_value(smargon.omega.max_velocity, 131)
    return RobotLoadThenCentreComposite(
        xbpm_feedback=xbpm_feedback,
        attenuator=attenuator,
        aperture_scatterguard=aperture_scatterguard,
        backlight=backlight,
        beamstop=beamstop_i03,
        detector_motion=detector_motion,
        eiger=eiger,
        zebra_fast_grid_scan=fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=pin_tip_detection_with_found_pin,
        smargon=smargon,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        zebra=zebra,
        zocalo=zocalo,
        panda=panda,
        panda_fast_grid_scan=panda_fast_grid_scan,
        thawer=thawer,
        sample_shutter=sample_shutter,
        vfm=vfm,
        mirror_voltages=mirror_voltages,
        dcm=dcm,
        undulator_dcm=undulator_dcm,
        robot=robot,
        webcam=webcam,
        lower_gonio=lower_gonio,
    )


@pytest.fixture
def robot_load_and_energy_change_composite(
    smargon,
    dcm,
    robot,
    aperture_scatterguard,
    oav,
    webcam,
    thawer,
    lower_gonio,
    vfm,
    mirror_voltages,
    undulator_dcm,
    xbpm_feedback,
    attenuator,
) -> RobotLoadAndEnergyChangeComposite:
    composite = RobotLoadAndEnergyChangeComposite(
        vfm,
        mirror_voltages,
        dcm,
        undulator_dcm,
        xbpm_feedback,
        attenuator,
        robot,
        webcam,
        lower_gonio,
        thawer,
        oav,
        smargon,
        aperture_scatterguard,
    )
    composite.smargon.stub_offsets.set = MagicMock(return_value=NullStatus())
    composite.aperture_scatterguard.set = MagicMock(return_value=NullStatus())
    set_mock_value(composite.dcm.energy_in_kev.user_readback, 11.105)

    return composite


def sim_fire_event_on_open_run(sim_run_engine: RunEngineSimulator, run_name: str):
    def fire_event(msg: Msg):
        try:
            sim_run_engine.fire_callback("start", msg.kwargs)
        except Exception as e:
            print(f"Exception is {e}")

    def msg_maches_run(msg: Msg):
        return msg.run == run_name

    sim_run_engine.add_handler("open_run", fire_event, msg_maches_run)


@pytest.fixture
def grid_detection_callback_with_detected_grid():
    with patch(
        "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.GridDetectionCallback",
        autospec=True,
    ) as callback:
        callback.return_value.get_grid_parameters.return_value = {
            "transmission_frac": 1.0,
            "exposure_time_s": 0,
            "x_start_um": 0,
            "y_start_um": 0,
            "y2_start_um": 0,
            "z_start_um": 0,
            "z2_start_um": 0,
            "x_steps": 10,
            "y_steps": 10,
            "z_steps": 10,
            "x_step_size_um": 0.1,
            "y_step_size_um": 0.1,
            "z_step_size_um": 0.1,
        }
        yield callback



from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.beamlines.i24.serial.parameters import get_chip_format, get_chip_map
from mx_bluesky.beamlines.i24.serial.parameters.utils import EmptyMapError


@pytest.mark.parametrize(
    "chip_type, expected_num_blocks, expected_step_size, expected_num_steps",
    [(0, 8, 0.125, 20), (1, 1, 0.6, 25), (3, 1, 0.125, 20)],
)
def test_get_chip_format_for_oxford_chips(
    chip_type: int,
    expected_num_blocks: int,
    expected_step_size: float,
    expected_num_steps: int,
):
    test_defaults = get_chip_format(ChipType(chip_type)).model_dump()

    assert (
        test_defaults["x_num_steps"] == expected_num_steps
        and test_defaults["x_num_steps"] == test_defaults["y_num_steps"]
    )
    assert test_defaults["y_blocks"] == expected_num_blocks
    assert test_defaults["x_step_size"] == expected_step_size


@patch("mx_bluesky.beamlines.i24.serial.parameters.utils.caget")
def test_get_chip_format_for_custom_chips(fake_caget: MagicMock):
    fake_caget.side_effect = ["10", "2", "0.2", "0.2"]
    test_chip_type = ChipType(2)
    test_defaults = get_chip_format(test_chip_type).model_dump()

    assert test_defaults["x_num_steps"] == 10
    assert test_defaults["y_num_steps"] == 2
    assert test_defaults["x_step_size"] == 0.2 and test_defaults["y_step_size"] == 0.2
    assert test_defaults["y_blocks"] == 1


@patch(
    "mx_bluesky.beamlines.i24.serial.parameters.utils.OXFORD_BLOCKS_PVS",
    new=["block1", "block2", "block3"],
)
@patch("mx_bluesky.beamlines.i24.serial.parameters.utils.caget")
def test_get_chip_map_raises_error_for_empty_map(fake_caget: MagicMock):
    fake_caget.side_effect = [0, 0, 0]
    with pytest.raises(EmptyMapError):
        get_chip_map()


@patch(
    "mx_bluesky.beamlines.i24.serial.parameters.utils.OXFORD_BLOCKS_PVS",
    new=["block1", "block2", "block3"],
)
@patch("mx_bluesky.beamlines.i24.serial.parameters.utils.caget")
def test_get_chip_map(fake_caget: MagicMock):
    fake_caget.side_effect = ["1", "0", "1"]

    chip_map = get_chip_map()

    assert len(chip_map) == 2
    assert chip_map == [1, 3]



import dataclasses
from collections.abc import Generator
from typing import cast
from unittest.mock import ANY, MagicMock, patch

import bluesky.plan_stubs as bps
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.backlight import BacklightPosition
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from ophyd_async.testing import get_mock_put, set_mock_value

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    _fire_xray_centre_result_event,
)
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
    detect_grid_and_do_gridscan,
    grid_detect_then_xray_centre,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)

from ....conftest import OavGridSnapshotTestEvents
from .conftest import FLYSCAN_RESULT_LOW, FLYSCAN_RESULT_MED, sim_fire_event_on_open_run


def _fake_flyscan(*args):
    yield from _fire_xray_centre_result_event([FLYSCAN_RESULT_MED, FLYSCAN_RESULT_LOW])


def test_full_grid_scan(
    test_fgs_params: HyperionSpecifiedThreeDGridScan, test_config_files: dict[str, str]
):
    devices = MagicMock()
    plan = grid_detect_then_xray_centre(
        devices,
        cast(GridScanWithEdgeDetect, test_fgs_params),
        test_config_files["oav_config_json"],
    )
    assert isinstance(plan, Generator)


@pytest.fixture
def grid_detect_devices_with_oav_config_params(
    grid_detect_devices: GridDetectThenXRayCentreComposite,
    test_config_files: dict[str, str],
) -> GridDetectThenXRayCentreComposite:
    set_mock_value(grid_detect_devices.oav.zoom_controller.level, "7.5x")
    return grid_detect_devices


@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.flyscan_xray_centre_no_move",
    autospec=True,
)
async def test_detect_grid_and_do_gridscan(
    mock_flyscan: MagicMock,
    pin_tip_detection_with_found_pin: PinTipDetection,
    grid_detect_devices_with_oav_config_params: GridDetectThenXRayCentreComposite,
    RE: RunEngine,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict,
):
    composite = grid_detect_devices_with_oav_config_params

    RE(
        ispyb_activation_wrapper(
            _do_detect_grid_and_gridscan_then_wait_for_backlight(
                composite, test_config_files, test_full_grid_scan_params
            ),
            test_full_grid_scan_params,
        )
    )

    # Check backlight was moved OUT
    get_mock_put(composite.backlight.position).assert_called_once_with(
        BacklightPosition.OUT, wait=ANY
    )

    # Check aperture was changed to SMALL
    assert (
        await composite.aperture_scatterguard.selected_aperture.get_value()
        == ApertureValue.SMALL
    )

    # Check we called out to underlying fast grid scan plan
    mock_flyscan.assert_called_once_with(ANY, ANY)


def _do_detect_grid_and_gridscan_then_wait_for_backlight(
    composite, test_config_files, test_full_grid_scan_params
):
    yield from detect_grid_and_do_gridscan(
        composite,
        parameters=test_full_grid_scan_params,
        oav_params=OAVParameters("xrayCentring", test_config_files["oav_config_json"]),
    )
    yield from bps.wait(CONST.WAIT.GRID_READY_FOR_DC)


@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.flyscan_xray_centre_no_move",
    autospec=True,
)
def test_when_full_grid_scan_run_then_parameters_sent_to_fgs_as_expected(
    mock_flyscan: MagicMock,
    grid_detect_devices_with_oav_config_params: GridDetectThenXRayCentreComposite,
    RE: RunEngine,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict,
    pin_tip_detection_with_found_pin: PinTipDetection,
):
    oav_params = OAVParameters("xrayCentring", test_config_files["oav_config_json"])

    RE(
        ispyb_activation_wrapper(
            detect_grid_and_do_gridscan(
                grid_detect_devices_with_oav_config_params,
                parameters=test_full_grid_scan_params,
                oav_params=oav_params,
            ),
            test_full_grid_scan_params,
        )
    )

    params: HyperionSpecifiedThreeDGridScan = mock_flyscan.call_args[0][1]

    assert params.detector_params.num_triggers == 180
    assert params.FGS_params.x_axis.full_steps == 15
    assert params.FGS_params.y_axis.end == pytest.approx(-0.0649, 0.001)

    # Parameters can be serialized
    params.model_dump_json()


@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.flyscan_xray_centre_no_move",
    autospec=True,
)
def test_detect_grid_and_do_gridscan_does_not_activate_ispyb_callback(
    mock_flyscan,
    mock_grid_detection_plan,
    grid_detect_devices_with_oav_config_params: GridDetectThenXRayCentreComposite,
    sim_run_engine: RunEngineSimulator,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict[str, str],
):
    mock_grid_detection_plan.return_value = iter([Msg("save_oav_grids")])
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_run_engine.add_callback_handler_for_multiple(
        "save_oav_grids",
        [
            [
                (
                    "descriptor",
                    OavGridSnapshotTestEvents.test_descriptor_document_oav_snapshot,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xy,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xz,  # type: ignore
                ),
            ]
        ],
    )

    msgs = sim_run_engine.simulate_plan(
        detect_grid_and_do_gridscan(
            grid_detect_devices_with_oav_config_params,
            test_full_grid_scan_params,
            OAVParameters("xrayCentring", test_config_files["oav_config_json"]),
        )
    )

    activations = [
        msg
        for msg in msgs
        if msg.command == "open_run"
        and "GridscanISPyBCallback" in msg.kwargs["activate_callbacks"]
    ]
    assert not activations


@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.change_aperture_then_move_to_xtal",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.flyscan_xray_centre_no_move",
    autospec=True,
    side_effect=_fake_flyscan,
)
def test_grid_detect_then_xray_centre_centres_on_the_first_flyscan_result(
    mock_flyscan: MagicMock,
    mock_change_aperture_then_move_to_xtal: MagicMock,
    grid_detect_devices_with_oav_config_params: GridDetectThenXRayCentreComposite,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict[str, str],
    pin_tip_detection_with_found_pin: PinTipDetection,
    RE: RunEngine,
):
    RE(
        grid_detect_then_xray_centre(
            grid_detect_devices_with_oav_config_params,
            test_full_grid_scan_params,
            test_config_files["oav_config_json"],
        )
    )
    mock_change_aperture_then_move_to_xtal.assert_called_once()

    assert (
        mock_change_aperture_then_move_to_xtal.mock_calls[0].args[0]
        == FLYSCAN_RESULT_MED
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan.flyscan_xray_centre_no_move",
    autospec=True,
)
def test_grid_detect_then_xray_centre_activates_ispyb_callback(
    mock_flyscan,
    mock_grid_detection_plan,
    sim_run_engine: RunEngineSimulator,
    grid_detect_devices_with_oav_config_params: GridDetectThenXRayCentreComposite,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict[str, str],
):
    mock_grid_detection_plan.return_value = iter(
        [
            Msg("save_oav_grids"),
            Msg(
                "open_run",
                run=CONST.PLAN.FLYSCAN_RESULTS,
                xray_centre_results=[dataclasses.asdict(FLYSCAN_RESULT_MED)],
            ),
        ]
    )

    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    sim_run_engine.add_callback_handler_for_multiple(
        "save_oav_grids",
        [
            [
                (
                    "descriptor",
                    OavGridSnapshotTestEvents.test_descriptor_document_oav_snapshot,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xy,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xz,  # type: ignore
                ),
            ]
        ],
    )
    msgs = sim_run_engine.simulate_plan(
        grid_detect_then_xray_centre(
            grid_detect_devices_with_oav_config_params,
            test_full_grid_scan_params,
            test_config_files["oav_config_json"],
        )
    )

    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and "GridscanISPyBCallback" in msg.kwargs["activate_callbacks"],
    )



import os
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from bluesky.plan_stubs import null
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.common.types import UpdatingPathProvider
from dodal.devices.fast_grid_scan import PandAGridScanParams
from dodal.devices.smargon import Smargon
from ophyd_async.fastcs.panda import HDFPanda, SeqTable, SeqTrigger

from mx_bluesky.common.parameters.constants import DeviceSettingsConstants
from mx_bluesky.hyperion.device_setup_plans.setup_panda import (
    MM_TO_ENCODER_COUNTS,
    PULSE_WIDTH_US,
    disarm_panda_for_gridscan,
    set_panda_directory,
    setup_panda_for_flyscan,
)


def get_smargon_speed(x_step_size_mm: float, time_between_x_steps_ms: float) -> float:
    return x_step_size_mm / time_between_x_steps_ms


def run_simulating_setup_panda_functions(
    plan: str,
    panda: HDFPanda,
    smargon: Smargon,
):
    num_of_sets = 0
    num_of_waits = 0

    def count_commands(msg):
        nonlocal num_of_sets
        nonlocal num_of_waits
        if msg.command == "set":
            num_of_sets += 1
        elif msg.command == "wait":
            num_of_waits += 1

    sim = RunEngineSimulator()
    sim.add_handler(["set", "wait"], count_commands)

    with patch(
        "mx_bluesky.hyperion.device_setup_plans.setup_panda.load_panda_from_yaml"
    ) as mock_load_panda:
        if plan == "setup":
            smargon_speed = get_smargon_speed(0.1, 1)
            sim.simulate_plan(
                setup_panda_for_flyscan(
                    panda,
                    PandAGridScanParams(transmission_fraction=0.01),
                    smargon,
                    0.1,
                    100.1,
                    smargon_speed,
                )
            )
            mock_load_panda.assert_called_once()
        elif plan == "disarm":
            sim.simulate_plan(disarm_panda_for_gridscan(panda))

    return num_of_sets, num_of_waits


def test_setup_panda_performs_correct_plans(sim_run_engine, panda, smargon):
    num_of_sets, num_of_waits = run_simulating_setup_panda_functions(
        "setup", panda, smargon
    )
    assert num_of_sets == 10
    assert num_of_waits == 5


@pytest.mark.parametrize(
    "x_steps, x_step_size, x_start, run_up_distance_mm, time_between_x_steps_ms, exposure_time_s",
    [
        (10, 0.2, 0, 0.5, 10.001, 0.01),
        (10, 0.5, -1, 0.05, 10.001, 0.01),
        (1, 2, 1.2, 1, 100.001, 0.1),
        (10, 2, -0.5, 3, 101, 0.1),
    ],
)
@patch("mx_bluesky.hyperion.device_setup_plans.setup_panda.load_panda_from_yaml")
def test_setup_panda_correctly_configures_table(
    mock_load_panda,
    x_steps: int,
    x_step_size: float,
    x_start: float,
    run_up_distance_mm: float,
    time_between_x_steps_ms: float,
    exposure_time_s: float,
    sim_run_engine: RunEngineSimulator,
    panda,
    smargon,
):
    sample_velocity_mm_per_s = get_smargon_speed(x_step_size, time_between_x_steps_ms)
    params = PandAGridScanParams(
        x_steps=x_steps,
        x_step_size_mm=x_step_size,
        x_start_mm=x_start,
        run_up_distance_mm=run_up_distance_mm,
        transmission_fraction=0.01,
    )

    exposure_distance_mm = sample_velocity_mm_per_s * exposure_time_s

    msgs = sim_run_engine.simulate_plan(
        setup_panda_for_flyscan(
            panda,
            params,
            smargon,
            exposure_time_s,
            time_between_x_steps_ms,
            sample_velocity_mm_per_s,
        )
    )

    # ignore all loading operations related to loading saved panda state from yaml
    msgs = [
        msg for msg in msgs if not msg.kwargs.get("group", "").startswith("load-phase")
    ]

    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "panda-pulse-1-width"
        and msg.args[0] == exposure_time_s,
    )

    table_msg = [
        msg
        for msg in msgs
        if msg.command == "set" and msg.obj.name == "panda-seq-1-table"
    ][0]

    table = table_msg.args[0]
    SPACE_WIDTH_US = int(time_between_x_steps_ms * 1000 - PULSE_WIDTH_US)

    exposure_distance_counts = exposure_distance_mm * MM_TO_ENCODER_COUNTS
    expected_seq_table: SeqTable = (
        SeqTable.row(
            repeats=1,
            trigger=SeqTrigger.BITA_1,
            position=0,
            time1=False,
            outa1=False,
            time2=True,
            outa2=False,
        )
        + SeqTable.row(
            repeats=x_steps,
            trigger=SeqTrigger.POSA_GT,
            position=int(params.x_start_mm * MM_TO_ENCODER_COUNTS),
            time1=PULSE_WIDTH_US,
            outa1=True,
            time2=SPACE_WIDTH_US,
            outa2=False,
        )
        + SeqTable.row(
            repeats=1,
            trigger=SeqTrigger.BITA_1,
            position=0,
            time1=False,
            outa1=False,
            time2=True,
            outa2=False,
        )
        + SeqTable.row(
            repeats=x_steps,
            trigger=SeqTrigger.POSA_LT,
            position=int(
                (params.x_start_mm + (params.x_steps - 1) * params.x_step_size_mm)
                * MM_TO_ENCODER_COUNTS
                + exposure_distance_counts
            ),
            time1=PULSE_WIDTH_US,
            outa1=True,
            time2=SPACE_WIDTH_US,
            outa2=False,
        )
    )

    for attr_name in table.__annotations__.keys():
        np.testing.assert_array_equal(
            getattr(table, attr_name), getattr(expected_seq_table, attr_name)
        )


def test_wait_between_setting_table_and_arming_panda(RE: RunEngine, panda, smargon):
    bps_wait_done = False

    def handle_wait(*args, **kwargs):
        nonlocal bps_wait_done
        bps_wait_done = True
        yield from null()

    def assert_set_table_has_been_waited_on(*args, **kwargs):
        assert bps_wait_done
        yield from null()

    with (
        patch(
            "mx_bluesky.hyperion.device_setup_plans.setup_panda.arm_panda_for_gridscan",
            MagicMock(side_effect=assert_set_table_has_been_waited_on),
        ),
        patch(
            "mx_bluesky.hyperion.device_setup_plans.setup_panda.bps.wait",
            MagicMock(side_effect=handle_wait),
        ),
        patch(
            "mx_bluesky.hyperion.device_setup_plans.setup_panda.load_panda_from_yaml"
        ),
        patch("mx_bluesky.hyperion.device_setup_plans.setup_panda.bps.abs_set"),
    ):
        RE(
            setup_panda_for_flyscan(
                panda,
                PandAGridScanParams(transmission_fraction=0.01),
                smargon,
                0.1,
                101.1,
                get_smargon_speed(0.1, 1),
            )
        )

    assert bps_wait_done


# It also would be useful to have some system tests which check that (at least)
# all the blocks which were enabled on setup are also disabled on tidyup
def test_disarm_panda_disables_correct_blocks(sim_run_engine, panda, smargon):
    num_of_sets, num_of_waits = run_simulating_setup_panda_functions(
        "disarm", panda, smargon
    )
    assert num_of_sets == 5
    assert num_of_waits == 1


@patch("mx_bluesky.hyperion.device_setup_plans.setup_panda.get_path_provider")
@patch("mx_bluesky.hyperion.device_setup_plans.setup_panda.datetime", spec=datetime)
def test_set_panda_directory(
    mock_datetime, mock_get_path_provider: MagicMock, tmp_path, RE
):
    mock_directory_provider = MagicMock(spec=UpdatingPathProvider)
    mock_datetime.now = MagicMock(
        return_value=datetime.fromisoformat("2024-08-11T15:59:23")
    )
    mock_get_path_provider.return_value = mock_directory_provider

    RE(set_panda_directory(tmp_path))
    mock_directory_provider.update.assert_called_with(
        directory=tmp_path, suffix="_20240811155923"
    )


def test_panda_settings_exist():
    # Panda settings are found relative to the top of the repo so this should
    # work in prod as well as test

    panda_settings_location = Path(
        DeviceSettingsConstants.PANDA_FLYSCAN_SETTINGS_DIR,
        f"{DeviceSettingsConstants.PANDA_FLYSCAN_SETTINGS_FILENAME}.yaml",
    )
    assert os.path.exists(panda_settings_location)



import dataclasses

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.run_engine import RunEngine

from mx_bluesky.common.external_interaction.callbacks.common.abstract_event import (
    AbstractEvent,
)


@dataclasses.dataclass(frozen=True)
class MyEvent(AbstractEvent):
    my_int: int
    my_float: float
    my_string: str
    my_bool: bool


def test_can_create_and_raise_an_event(RE: RunEngine):
    @bpp.run_decorator()
    def fire_an_event():
        yield from bps.create("MY_EVENT")
        yield from bps.read(
            MyEvent(my_int=123, my_float=1.234, my_string="Test message", my_bool=True)
        )
        yield from bps.save()

    the_event = None

    def my_event_handler(name: str, doc: dict):
        nonlocal the_event
        the_event = doc

    RE.subscribe(my_event_handler, "event")
    RE(fire_an_event())

    assert the_event
    assert the_event["data"] == {
        "my_bool": True,
        "my_float": 1.234,
        "my_string": "Test message",
        "my_int": 123,
    }



import re
from collections.abc import Callable, Generator
from typing import TypeVar

from bluesky.plan_stubs import null
from bluesky.preprocessors import contingency_wrapper
from bluesky.utils import Msg


class WarningException(Exception):
    """An exception used when we want to warn GDA of a
    problem but continue with UDC anyway"""

    pass


class ISPyBDepositionNotMade(Exception):
    """Raised when the ISPyB or Zocalo callbacks can't access ISPyB deposition numbers."""

    pass


class SampleException(WarningException):
    """An exception which identifies an issue relating to the sample."""

    def __str__(self):
        class_name = type(self).__name__
        return f"[{class_name}]: {super().__str__()}"

    @classmethod
    def type_and_message_from_reason(cls, reason: str) -> tuple[str, str]:
        match = re.match(r"\[(\S*)?]: (.*)", reason)
        return (match.group(1), match.group(2)) if match else (None, None)


T = TypeVar("T")


class CrystalNotFoundException(SampleException):
    """Raised if grid detection completed normally but no crystal was found."""

    def __init__(self, *args):
        super().__init__("Diffraction not found, skipping sample.")


def catch_exception_and_warn(
    exception_to_catch: type[Exception],
    func: Callable[..., Generator[Msg, None, T]],
    *args,
    **kwargs,
) -> Generator[Msg, None, T]:
    """A plan wrapper to catch a specific exception and instead raise a WarningException,
    so that UDC is not halted

    Example usage:

    'def plan_which_can_raise_exception_a(*args, **kwargs):
        ...
    yield from catch_exception_and_warn(ExceptionA, plan_which_can_raise_exception_a, **args, **kwargs)'

    This will catch ExceptionA raised by the plan and instead raise a WarningException
    """

    def warn_if_exception_matches(exception: Exception):
        if isinstance(exception, exception_to_catch):
            raise SampleException(str(exception)) from exception
        yield from null()

    return (
        yield from contingency_wrapper(
            func(*args, **kwargs),
            except_plan=warn_if_exception_matches,
        )
    )



from dataclasses import replace
from functools import partial
from itertools import dropwhile
from unittest.mock import MagicMock, patch

import pytest
from ispyb.sp.mxacquisition import MXAcquisition

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionGroupInfo,
    DataCollectionInfo,
    DataCollectionPositionInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)

from ......conftest import (
    EXPECTED_END_TIME,
    EXPECTED_START_TIME,
    TEST_BARCODE,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
    TEST_GRID_INFO_IDS,
    TEST_SAMPLE_ID,
    TEST_SESSION_ID,
    assert_upsert_call_with,
    mx_acquisition_from_conn,
    remap_upsert_columns,
)


@pytest.fixture
def dummy_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="Mesh3D",
        sample_id=364758,
    )


DC_INFO_FOR_BEGIN_XY = DataCollectionInfo(
    omega_start=0.0,
    data_collection_number=1,
    xtal_snapshot1="test_1_y",
    xtal_snapshot2="test_2_y",
    xtal_snapshot3="test_3_y",
    n_images=800,
    axis_range=0,
    axis_end=0.0,
    kappa_start=None,
    parent_id=None,
    visit_string="cm31105-4",
    sample_id=364758,
    detector_id=78,
    axis_start=0.0,
    focal_spot_size_at_samplex=0.0,
    focal_spot_size_at_sampley=0.0,
    slitgap_vertical=0.1,
    slitgap_horizontal=0.1,
    beamsize_at_samplex=0.1,
    beamsize_at_sampley=0.1,
    transmission=100.0,
    comments="MX-Bluesky: Xray centring 1 -",
    detector_distance=100.0,
    exp_time=0.1,
    imgdir="/tmp/",
    file_template="file_name_0_master.h5",
    imgprefix="file_name",
    imgsuffix="h5",
    n_passes=1,
    overlap=0,
    start_image_number=1,
    wavelength=123.98419840550369,
    xbeam=150.0,
    ybeam=160.0,
    synchrotron_mode=None,
    undulator_gap1=1.0,
    start_time=EXPECTED_START_TIME,
)

DC_INFO_FOR_BEGIN_XZ = replace(
    DC_INFO_FOR_BEGIN_XY,
    xtal_snapshot1="test_1_z",
    xtal_snapshot2="test_2_z",
    xtal_snapshot3="test_3_z",
    omega_start=90.0,
    n_images=400,
    axis_end=90.0,
    axis_start=90.0,
    file_template="file_name_1_master.h5",
    comments="MX-Bluesky: Xray centring 2 -",
)

DC_INFO_FOR_UPDATE_XY = replace(
    DC_INFO_FOR_BEGIN_XY,
    parent_id=34,
    comments="Diffraction grid scan of 40 by 20 images in 100.0 um by 100.0 um steps. Top left (px): [50,100], bottom right (px): [3250,1700].",
    flux=10.0,
    synchrotron_mode="test",
)

DC_INFO_FOR_UPDATE_XZ = replace(
    DC_INFO_FOR_BEGIN_XZ,
    parent_id=34,
    comments="Diffraction grid scan of 40 by 10 images in 100.0 um by 200.0 um steps. Top left (px): [50,120], bottom right (px): [3250,1720].",
    flux=10.0,
    synchrotron_mode="test",
)

EXPECTED_BASE_UPSERT = {
    "visitid": TEST_SESSION_ID,
    "parentid": TEST_DATA_COLLECTION_GROUP_ID,
    "sampleid": TEST_SAMPLE_ID,
    "detectorid": 78,
    "axisrange": 0,
    "focal_spot_size_at_samplex": 0.0,
    "focal_spot_size_at_sampley": 0.0,
    "slitgap_vertical": 0.1,
    "slitgap_horizontal": 0.1,
    "beamsize_at_samplex": 0.1,
    "beamsize_at_sampley": 0.1,
    "transmission": 100.0,
    "data_collection_number": 1,
    "detector_distance": 100.0,
    "exp_time": 0.1,
    "imgdir": "/tmp/",
    "imgprefix": "file_name",
    "imgsuffix": "h5",
    "n_passes": 1,
    "overlap": 0,
    "start_image_number": 1,
    "wavelength": 123.98419840550369,
    "xbeam": 150.0,
    "ybeam": 160.0,
    "undulator_gap1": 1.0,
    "starttime": EXPECTED_START_TIME,
}

EXPECTED_BASE_XY_UPSERT = EXPECTED_BASE_UPSERT | {
    "xtal_snapshot1": "test_1_y",
    "xtal_snapshot2": "test_2_y",
    "xtal_snapshot3": "test_3_y",
    "omegastart": 0,
    "axisstart": 0.0,
    "axisend": 0,
    "filetemplate": "file_name_0_master.h5",
    "nimages": 40 * 20,
}

EXPECTED_BASE_XZ_UPSERT = EXPECTED_BASE_UPSERT | {
    "xtal_snapshot1": "test_1_z",
    "xtal_snapshot2": "test_2_z",
    "xtal_snapshot3": "test_3_z",
    "omegastart": 90.0,
    "axisstart": 90.0,
    "axisend": 90.0,
    "filetemplate": "file_name_1_master.h5",
    "nimages": 40 * 10,
}

EXPECTED_DC_XY_BEGIN_UPSERT = EXPECTED_BASE_XY_UPSERT | {
    "comments": "MX-Bluesky: Xray centring 1 -",
}

EXPECTED_DC_XZ_BEGIN_UPSERT = EXPECTED_BASE_XZ_UPSERT | {
    "comments": "MX-Bluesky: Xray centring 2 -",
}

EXPECTED_DC_XY_UPDATE_UPSERT = EXPECTED_BASE_XY_UPSERT | {
    "id": 12,
    "flux": 10.0,
    "synchrotron_mode": "test",
}

EXPECTED_DC_XZ_UPDATE_UPSERT = EXPECTED_BASE_XZ_UPSERT | {
    "id": 13,
    "flux": 10,
    "synchrotron_mode": "test",
}


@pytest.fixture
def scan_data_infos_for_begin():
    return [
        ScanDataInfo(
            data_collection_info=replace(DC_INFO_FOR_BEGIN_XY),
            data_collection_id=None,
            data_collection_position_info=None,
            data_collection_grid_info=None,
        ),
        ScanDataInfo(
            data_collection_info=replace(DC_INFO_FOR_BEGIN_XZ),
            data_collection_id=None,
            data_collection_position_info=None,
            data_collection_grid_info=None,
        ),
    ]


@pytest.fixture
def scan_data_infos_for_update():
    scan_xy_data_info_for_update = ScanDataInfo(
        data_collection_info=replace(DC_INFO_FOR_UPDATE_XY),
        data_collection_id=TEST_DATA_COLLECTION_IDS[0],
        data_collection_position_info=DataCollectionPositionInfo(
            pos_x=0, pos_y=0, pos_z=0
        ),
        data_collection_grid_info=DataCollectionGridInfo(
            dx_in_mm=0.1,
            dy_in_mm=0.1,
            steps_x=40,
            steps_y=20,
            microns_per_pixel_x=1.25,
            microns_per_pixel_y=1.25,
            snapshot_offset_x_pixel=50,
            snapshot_offset_y_pixel=100,
            orientation=Orientation.HORIZONTAL,
            snaked=True,
        ),
    )
    scan_xz_data_info_for_update = ScanDataInfo(
        data_collection_info=replace(DC_INFO_FOR_UPDATE_XZ),
        data_collection_id=TEST_DATA_COLLECTION_IDS[1],
        data_collection_position_info=DataCollectionPositionInfo(
            pos_x=0.0, pos_y=0.0, pos_z=0.0
        ),
        data_collection_grid_info=DataCollectionGridInfo(
            dx_in_mm=0.1,
            dy_in_mm=0.2,
            steps_x=40,
            steps_y=10,
            microns_per_pixel_x=1.25,
            microns_per_pixel_y=1.25,
            snapshot_offset_x_pixel=50,
            snapshot_offset_y_pixel=120,
            orientation=Orientation.HORIZONTAL,
            snaked=True,
        ),
    )
    return [scan_xy_data_info_for_update, scan_xz_data_info_for_update]


def test_ispyb_deposition_comment_for_3D_correct(
    mock_ispyb_conn: MagicMock,
    dummy_3d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    mock_ispyb_conn = mock_ispyb_conn
    mock_mx_aquisition = mx_acquisition_from_conn(mock_ispyb_conn)
    mock_upsert_dc = mock_mx_aquisition.upsert_data_collection
    ispyb_ids = dummy_3d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_3d_gridscan_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    upsert_keys = mock_mx_aquisition.get_data_collection_params()
    first_upserted_param_value_dict = remap_upsert_columns(
        upsert_keys, mock_upsert_dc.call_args_list[0][0][0]
    )
    second_upserted_param_value_dict = remap_upsert_columns(
        upsert_keys, mock_upsert_dc.call_args_list[1][0][0]
    )
    assert first_upserted_param_value_dict["comments"] == (
        "MX-Bluesky: Xray centring 1 -"
    )
    assert second_upserted_param_value_dict["comments"] == (
        "MX-Bluesky: Xray centring 2 -"
    )
    mock_mx_aquisition.update_data_collection_append_comments.assert_any_call(
        TEST_DATA_COLLECTION_IDS[0],
        "Diffraction grid scan of 40 by 20 images "
        "in 100.0 um by 100.0 um steps. Top left (px): [50,100], bottom right (px): [3250,1700].",
        " ",
    )
    mock_mx_aquisition.update_data_collection_append_comments.assert_any_call(
        TEST_DATA_COLLECTION_IDS[1],
        "Diffraction grid scan of 40 by 10 images "
        "in 100.0 um by 200.0 um steps. Top left (px): [50,120], bottom right (px): [3250,1720].",
        " ",
    )


def test_store_3d_grid_scan(
    mock_ispyb_conn,
    dummy_3d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_3d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    assert ispyb_ids == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0], TEST_DATA_COLLECTION_IDS[1]),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )

    assert dummy_3d_gridscan_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    ) == IspybIds(
        data_collection_ids=TEST_DATA_COLLECTION_IDS,
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        grid_ids=TEST_GRID_INFO_IDS,
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition(
    mock_ispyb_conn,
    dummy_3d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
):
    assert dummy_3d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0], TEST_DATA_COLLECTION_IDS[1]),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )

    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    assert_upsert_call_with(
        mx_acq.upsert_data_collection_group.mock_calls[0],
        mx_acq.get_data_collection_group_params(),
        {
            "parentid": TEST_SESSION_ID,
            "experimenttype": "Mesh3D",
            "sampleid": TEST_SAMPLE_ID,
        },
    )
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        EXPECTED_DC_XY_BEGIN_UPSERT,
    )
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[1],
        mx_acq.get_data_collection_params(),
        EXPECTED_DC_XZ_BEGIN_UPSERT,
    )
    mx_acq.update_dc_position.assert_not_called()
    mx_acq.upsert_dc_grid.assert_not_called()


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_update_deposition(
    mock_ispyb_conn,
    dummy_3d_gridscan_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_3d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    mx_acq.upsert_data_collection_group.assert_called_once()
    mx_acq.upsert_data_collection.assert_called()
    mx_acq.upsert_data_collection_group.reset_mock()
    mx_acq.upsert_data_collection.reset_mock()

    dummy_collection_group_info.sample_barcode = TEST_BARCODE

    actual_rows = dummy_3d_gridscan_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    )

    assert actual_rows == IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        data_collection_ids=TEST_DATA_COLLECTION_IDS,
        grid_ids=TEST_GRID_INFO_IDS,
    )

    mx_acq.upsert_data_collection_group.assert_not_called()

    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[0],
        mx_acq.get_data_collection_params(),
        EXPECTED_DC_XY_UPDATE_UPSERT,
    )

    assert mx_acq.update_data_collection_append_comments.call_args_list[0] == (
        (
            TEST_DATA_COLLECTION_IDS[0],
            "Diffraction grid scan of 40 by 20 "
            "images in 100.0 um by 100.0 um steps. Top left (px): [50,100], "
            "bottom right (px): [3250,1700].",
            " ",
        ),
    )

    assert_upsert_call_with(
        mx_acq.update_dc_position.mock_calls[0],
        mx_acq.get_dc_position_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "pos_x": 0,
            "pos_y": 0,
            "pos_z": 0,
        },
    )

    assert_upsert_call_with(
        mx_acq.upsert_dc_grid.mock_calls[0],
        mx_acq.get_dc_grid_params(),
        {
            "parentid": TEST_DATA_COLLECTION_IDS[0],
            "dxinmm": 0.1,
            "dyinmm": 0.1,
            "stepsx": 40,
            "stepsy": 20,
            "micronsperpixelx": 1.25,
            "micronsperpixely": 1.25,
            "snapshotoffsetxpixel": 50,
            "snapshotoffsetypixel": 100,
            "orientation": "horizontal",
            "snaked": True,
        },
    )

    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[1],
        mx_acq.get_data_collection_params(),
        EXPECTED_DC_XZ_UPDATE_UPSERT,
    )

    assert mx_acq.update_data_collection_append_comments.call_args_list[1] == (
        (
            TEST_DATA_COLLECTION_IDS[1],
            "Diffraction grid scan of 40 by 10 "
            "images in 100.0 um by 200.0 um steps. Top left (px): [50,120], "
            "bottom right (px): [3250,1720].",
            " ",
        ),
    )

    assert_upsert_call_with(
        mx_acq.update_dc_position.mock_calls[1],
        mx_acq.get_dc_position_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[1],
            "pos_x": 0,
            "pos_y": 0,
            "pos_z": 0,
        },
    )

    assert_upsert_call_with(
        mx_acq.upsert_dc_grid.mock_calls[1],
        mx_acq.get_dc_grid_params(),
        {
            "parentid": TEST_DATA_COLLECTION_IDS[1],
            "dxinmm": 0.1,
            "dyinmm": 0.2,
            "stepsx": 40,
            "stepsy": 10,
            "micronsperpixelx": 1.25,
            "micronsperpixely": 1.25,
            "snapshotoffsetxpixel": 50,
            "snapshotoffsetypixel": 120,
            "orientation": "horizontal",
            "snaked": True,
        },
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
@patch(
    "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
)
def test_end_deposition_happy_path(
    get_current_time,
    mock_ispyb_conn,
    dummy_3d_gridscan_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_3d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    assert len(mx_acq.upsert_data_collection_group.mock_calls) == 1
    ispyb_ids = dummy_3d_gridscan_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    )
    assert len(mx_acq.upsert_data_collection_group.mock_calls) == 1
    assert len(mx_acq.upsert_data_collection.mock_calls) == 4
    assert len(mx_acq.upsert_dc_grid.mock_calls) == 2

    get_current_time.return_value = EXPECTED_END_TIME
    dummy_3d_gridscan_ispyb.end_deposition(ispyb_ids, "success", "Test succeeded")
    mx_acq.update_data_collection_append_comments.assert_any_call(
        TEST_DATA_COLLECTION_IDS[0],
        "DataCollection Successful reason: Test succeeded",
        " ",
    )
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[4],
        mx_acq.get_data_collection_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[0],
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "endtime": EXPECTED_END_TIME,
            "runstatus": "DataCollection Successful",
        },
    )
    mx_acq.update_data_collection_append_comments.assert_any_call(
        TEST_DATA_COLLECTION_IDS[1],
        "DataCollection Successful reason: Test succeeded",
        " ",
    )
    assert_upsert_call_with(
        mx_acq.upsert_data_collection.mock_calls[5],
        mx_acq.get_data_collection_params(),
        {
            "id": TEST_DATA_COLLECTION_IDS[1],
            "parentid": TEST_DATA_COLLECTION_GROUP_ID,
            "endtime": EXPECTED_END_TIME,
            "runstatus": "DataCollection Successful",
        },
    )


def test_param_keys(
    mock_ispyb_conn,
    dummy_2d_gridscan_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    ispyb_ids = dummy_2d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    assert dummy_2d_gridscan_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0], TEST_DATA_COLLECTION_IDS[1]),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        grid_ids=(TEST_GRID_INFO_IDS[0], TEST_GRID_INFO_IDS[1]),
    )


def _test_when_grid_scan_stored_then_data_present_in_upserts(
    ispyb_conn,
    dummy_ispyb,
    test_function,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
    test_group=False,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    mx_acquisition = mx_acquisition_from_conn(ispyb_conn)

    def call_does_not_have_dcid(id, func_call):
        return func_call.args[0][0] != id

    for dc_id in ispyb_ids.data_collection_ids:
        upsert_call = next(
            dropwhile(
                partial(call_does_not_have_dcid, dc_id),
                mx_acquisition.upsert_data_collection.call_args_list,
            )
        )
        actual = upsert_call[0][0]
        assert test_function(MXAcquisition.get_data_collection_params(), actual)

    if test_group:
        upsert_data_collection_group_arg_list = (
            mx_acquisition.upsert_data_collection_group.call_args_list[0][0]
        )
        actual = upsert_data_collection_group_arg_list[0]
        assert test_function(MXAcquisition.get_data_collection_group_params(), actual)


def test_given_sampleid_of_none_when_grid_scan_stored_then_sample_id_not_set(
    mock_ispyb_conn,
    dummy_2d_gridscan_ispyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    dummy_collection_group_info.sample_id = None
    for dc_info in [
        scan_info.data_collection_info
        for scan_info in scan_data_infos_for_begin + scan_data_infos_for_update
    ]:
        dc_info.sample_id = None

    def test_sample_id(default_params, actual):
        sampleid_idx = list(default_params).index("sampleid")
        return actual[sampleid_idx] == default_params["sampleid"]

    _test_when_grid_scan_stored_then_data_present_in_upserts(
        mock_ispyb_conn,
        dummy_2d_gridscan_ispyb,
        test_sample_id,
        dummy_collection_group_info,
        scan_data_infos_for_begin,
        scan_data_infos_for_update,
        True,
    )


def test_given_real_sampleid_when_grid_scan_stored_then_sample_id_set(
    mock_ispyb_conn,
    dummy_2d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    expected_sample_id = 364758

    def test_sample_id(default_params, actual):
        sampleid_idx = list(default_params).index("sampleid")
        return actual[sampleid_idx] == expected_sample_id

    _test_when_grid_scan_stored_then_data_present_in_upserts(
        mock_ispyb_conn,
        dummy_2d_gridscan_ispyb,
        test_sample_id,
        dummy_collection_group_info,
        scan_data_infos_for_begin,
        scan_data_infos_for_update,
        True,
    )


def test_fail_result_run_results_in_bad_run_status(
    mock_ispyb_conn: MagicMock,
    dummy_2d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    mock_ispyb_conn = mock_ispyb_conn
    mock_mx_aquisition = (
        mock_ispyb_conn.return_value.__enter__.return_value.mx_acquisition
    )
    mock_upsert_data_collection = mock_mx_aquisition.upsert_data_collection

    ispyb_ids = dummy_2d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    ispyb_ids = dummy_2d_gridscan_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    )
    dummy_2d_gridscan_ispyb.end_deposition(ispyb_ids, "fail", "test specifies failure")

    mock_upsert_data_collection_calls = mock_upsert_data_collection.call_args_list
    for upsert_call in mock_upsert_data_collection_calls[4:5]:
        end_deposition_upsert_args = upsert_call[0]
        upserted_param_value_list = end_deposition_upsert_args[0]
        assert "DataCollection Unsuccessful" in upserted_param_value_list
        assert "DataCollection Successful" not in upserted_param_value_list


def test_no_exception_during_run_results_in_good_run_status(
    mock_ispyb_conn: MagicMock,
    dummy_2d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info,
    scan_data_infos_for_begin,
    scan_data_infos_for_update,
):
    mock_mx_acquisition = (
        mock_ispyb_conn.return_value.__enter__.return_value.mx_acquisition
    )
    mock_upsert_data_collection = mock_mx_acquisition.upsert_data_collection

    ispyb_ids = dummy_2d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    ispyb_ids = dummy_2d_gridscan_ispyb.update_deposition(
        ispyb_ids, scan_data_infos_for_update
    )
    dummy_2d_gridscan_ispyb.end_deposition(ispyb_ids, "success", "")

    mock_upsert_data_collection_calls = mock_upsert_data_collection.call_args_list
    for upsert_call in mock_upsert_data_collection_calls[4:5]:
        end_deposition_upsert_args = upsert_call[0]
        upserted_param_value_list = end_deposition_upsert_args[0]
        assert "DataCollection Unsuccessful" not in upserted_param_value_list
        assert "DataCollection Successful" in upserted_param_value_list


def test_update_data_collection_no_comment(
    mock_ispyb_conn: MagicMock,
    dummy_3d_gridscan_ispyb: StoreInIspyb,
    dummy_collection_group_info: DataCollectionGroupInfo,
    scan_data_infos_for_begin: list[ScanDataInfo],
    scan_data_infos_for_update: list[ScanDataInfo],
):
    for scan_data_info in scan_data_infos_for_update:
        scan_data_info.data_collection_info.comments = None

    ispyb_ids = dummy_3d_gridscan_ispyb.begin_deposition(
        dummy_collection_group_info, scan_data_infos_for_begin
    )
    dummy_3d_gridscan_ispyb.update_deposition(ispyb_ids, scan_data_infos_for_update)

    mx_acq = mx_acquisition_from_conn(mock_ispyb_conn)
    mx_acq.update_data_collection_append_comments.assert_not_called()



from collections.abc import Generator

import bluesky.plan_stubs as bps
import pydantic
from blueapi.core import BlueskyContext
from bluesky.utils import Msg
from dodal.devices.backlight import Backlight
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAV_CONFIG_JSON, OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection, Tip
from dodal.devices.oav.utils import (
    PinNotFoundException,
    Pixel,
    get_move_required_so_that_beam_is_at_pixel,
    wait_for_tip_to_be_found,
)
from dodal.devices.smargon import Smargon

from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.exceptions import SampleException, catch_exception_and_warn
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.device_setup_plans.setup_oav import pre_centring_setup_oav
from mx_bluesky.hyperion.device_setup_plans.smargon import (
    move_smargon_warn_on_out_of_range,
)
from mx_bluesky.hyperion.parameters.constants import CONST

DEFAULT_STEP_SIZE = 0.5


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class PinTipCentringComposite:
    """All devices which are directly or indirectly required by this plan"""

    backlight: Backlight
    oav: OAV
    smargon: Smargon
    pin_tip_detection: PinTipDetection


def create_devices(context: BlueskyContext) -> PinTipCentringComposite:
    return device_composite_from_context(context, PinTipCentringComposite)


def trigger_and_return_pin_tip(
    pin_tip: PinTipDetection,
) -> Generator[Msg, None, Tip]:
    yield from bps.trigger(pin_tip, wait=True)
    tip_x_y_px = yield from bps.rd(pin_tip.triggered_tip)
    LOGGER.info(f"Pin tip found at {tip_x_y_px}")
    return tip_x_y_px


def move_pin_into_view(
    pin_tip_device: PinTipDetection,
    smargon: Smargon,
    step_magnitude_mm: float = DEFAULT_STEP_SIZE,
    max_steps: int = 2,
) -> Generator[Msg, None, Pixel]:
    """Attempt to move the pin into view and return the tip location in pixels if found.
    The gonio x is moved in a number of discrete steps to find the pin. If the move
    would take it past its limit, it moves to the limit instead.

    Args:
        pin_tip_device (PinTipDetection): The device being used to detect the pin
        smargon (Smargon): The gonio to move the tip
        step_magnitude_mm (float, optional): Distance to move the gonio (in mm) for each
                                    step of the search. Defaults to 0.5.
        max_steps (int, optional): The number of steps to search with. Defaults to 2.

    Raises:
        SampleException: Error if the pin tip is never found

    Returns:
        Tuple[int, int]: The location of the pin tip in pixels
    """

    def pin_tip_valid(pin_xy: Tip):
        return not all(pin_xy == pin_tip_device.INVALID_POSITION) and pin_xy[0] != 0

    for _ in range(max_steps):
        tip_xy_px = yield from trigger_and_return_pin_tip(pin_tip_device)

        if pin_tip_valid(tip_xy_px):
            return (int(tip_xy_px[0]), int(tip_xy_px[1]))

        # Pin is off in the -ve direction if the returned tip x pixel value is 0
        direction_multiple = -1 if tip_xy_px[0] == 0 else 1
        step_vector_mm = step_magnitude_mm * direction_multiple

        smargon_x = yield from bps.rd(smargon.x.user_readback)
        ideal_move_to_find_pin = float(smargon_x) + step_vector_mm
        high_limit = yield from bps.rd(smargon.x.high_limit_travel)
        low_limit = yield from bps.rd(smargon.x.low_limit_travel)
        move_within_limits = max(min(ideal_move_to_find_pin, high_limit), low_limit)
        if move_within_limits != ideal_move_to_find_pin:
            LOGGER.warning(
                f"Pin tip is off screen, and moving {step_vector_mm}mm would cross limits, "
                f"moving to {move_within_limits} instead"
            )
        yield from bps.mv(smargon.x, move_within_limits)

        # Some time for the view to settle after the move
        yield from bps.sleep(CONST.HARDWARE.OAV_REFRESH_DELAY)

    tip_xy_px = yield from trigger_and_return_pin_tip(pin_tip_device)

    if not pin_tip_valid(tip_xy_px):
        raise SampleException(
            "Pin tip centring failed - pin too long/short/bent and out of range"
        )
    else:
        return (int(tip_xy_px[0]), int(tip_xy_px[1]))


def pin_tip_centre_plan(
    composite: PinTipCentringComposite,
    tip_offset_microns: float,
    oav_config_file: str = OAV_CONFIG_JSON,
):
    """Finds the tip of the pin and moves to roughly the centre based on this tip. Does
    this at both the current omega angle and +90 deg from this angle so as to get a
    centre in 3D.

    Args:
        tip_offset_microns (float): The x offset from the tip where the centre is assumed
                                    to be.
    """
    oav: OAV = composite.oav
    smargon: Smargon = composite.smargon
    oav_params = OAVParameters("pinTipCentring", oav_config_file)

    pin_tip_setup = composite.pin_tip_detection
    pin_tip_detect = composite.pin_tip_detection

    microns_per_pixel_x = yield from bps.rd(oav.microns_per_pixel_x)
    tip_offset_px = int(tip_offset_microns / microns_per_pixel_x)

    def offset_and_move(tip: Pixel):
        pixel_to_move_to = (tip[0] + tip_offset_px, tip[1])
        position_mm = yield from get_move_required_so_that_beam_is_at_pixel(
            smargon, pixel_to_move_to, oav
        )
        LOGGER.info(f"Tip centring moving to : {position_mm}")
        yield from move_smargon_warn_on_out_of_range(smargon, position_mm)

    LOGGER.info(f"Tip offset in pixels: {tip_offset_px}")

    # need to wait for the OAV image to update
    # See #673 for improvements
    yield from bps.sleep(0.3)

    yield from pre_centring_setup_oav(oav, oav_params, pin_tip_setup)

    tip = yield from move_pin_into_view(pin_tip_detect, smargon)
    yield from offset_and_move(tip)

    yield from bps.mvr(smargon.omega, 90)

    # need to wait for the OAV image to update
    # See #673 for improvements
    yield from bps.sleep(0.3)
    tip = yield from catch_exception_and_warn(
        PinNotFoundException, wait_for_tip_to_be_found, pin_tip_detect
    )
    yield from offset_and_move(tip)



import json

import bluesky.plan_stubs as bps
from dodal.devices.focusing_mirror import (
    FocusingMirrorWithStripes,
    MirrorStripe,
    MirrorVoltages,
)
from dodal.devices.undulator_dcm import UndulatorDCM
from dodal.devices.util.adjuster_plans import lookup_table_adjuster
from dodal.devices.util.lookup_tables import (
    linear_interpolation_lut,
    parse_lookup_table,
)

from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.utils import (
    energy_to_bragg_angle,
)

MIRROR_VOLTAGE_GROUP = "MIRROR_VOLTAGE_GROUP"
DCM_GROUP = "DCM_GROUP"
YAW_LAT_TIMEOUT_S = 30


def _apply_and_wait_for_voltages_to_settle(
    stripe: MirrorStripe,
    mirror_voltages: MirrorVoltages,
):
    with open(mirror_voltages.voltage_lookup_table_path) as lut_file:
        json_obj = json.load(lut_file)

    # sample mode is the only mode supported
    sample_data = json_obj["sample"]
    if stripe == MirrorStripe.BARE:
        stripe_key = "bare"
    elif stripe == MirrorStripe.RHODIUM:
        stripe_key = "rh"
    elif stripe == MirrorStripe.PLATINUM:
        stripe_key = "pt"

    for mirror_key, channels in {
        "hfm": mirror_voltages.horizontal_voltages,
        "vfm": mirror_voltages.vertical_voltages,
    }.items():
        required_voltages = sample_data[stripe_key][mirror_key]

        for voltage_channel, required_voltage in zip(
            channels.values(), required_voltages, strict=True
        ):
            LOGGER.info(
                f"Applying and waiting for voltage {voltage_channel.name} = {required_voltage}"
            )
            yield from bps.abs_set(
                voltage_channel, required_voltage, group=MIRROR_VOLTAGE_GROUP, wait=True
            )


def adjust_mirror_stripe(
    energy_kev, mirror: FocusingMirrorWithStripes, mirror_voltages: MirrorVoltages
):
    """Feedback should be OFF prior to entry, in order to prevent
    feedback from making unnecessary corrections while beam is being adjusted."""
    mirror_config = mirror.energy_to_stripe(energy_kev)

    LOGGER.info(
        f"Adjusting mirror stripe for {energy_kev}keV selecting {mirror_config['stripe']} stripe"
    )
    yield from bps.abs_set(mirror.stripe, mirror_config["stripe"], wait=True)
    yield from bps.trigger(mirror.apply_stripe)

    # yaw, lat cannot be done simultaneously
    LOGGER.info(f"Adjusting {mirror.name} lat to {mirror_config['lat_mm']}")
    yield from bps.abs_set(
        mirror.x_mm, mirror_config["lat_mm"], wait=True, timeout=YAW_LAT_TIMEOUT_S
    )

    LOGGER.info(f"Adjusting {mirror.name} yaw to {mirror_config['yaw_mrad']}")
    yield from bps.abs_set(
        mirror.yaw_mrad, mirror_config["yaw_mrad"], wait=True, timeout=YAW_LAT_TIMEOUT_S
    )

    LOGGER.info("Adjusting mirror voltages...")
    yield from _apply_and_wait_for_voltages_to_settle(
        mirror_config["stripe"], mirror_voltages
    )


def adjust_dcm_pitch_roll_vfm_from_lut(
    undulator_dcm: UndulatorDCM,
    vfm: FocusingMirrorWithStripes,
    mirror_voltages: MirrorVoltages,
    energy_kev,
):
    """Beamline energy-change post-adjustments : Adjust DCM and VFM directly from lookup tables.
    Lookups are performed against the Bragg angle which is computed directly from the target energy
    rather than waiting for the EPICS controls PV to reach it.
    Feedback should be OFF prior to entry, in order to prevent
    feedback from making unnecessary corrections while beam is being adjusted."""

    # Adjust DCM Pitch
    dcm = undulator_dcm.dcm_ref()
    LOGGER.info(f"Adjusting DCM and VFM for {energy_kev} keV")
    d_spacing_a: float = yield from bps.rd(
        undulator_dcm.dcm_ref().crystal_metadata_d_spacing
    )
    bragg_deg = energy_to_bragg_angle(energy_kev, d_spacing_a)
    LOGGER.info(f"Target Bragg angle = {bragg_deg} degrees")
    dcm_pitch_adjuster = lookup_table_adjuster(
        linear_interpolation_lut(
            *parse_lookup_table(undulator_dcm.pitch_energy_table_path)
        ),
        dcm.pitch_in_mrad,
        bragg_deg,
    )
    yield from dcm_pitch_adjuster(DCM_GROUP)
    # It's possible we can remove these waits but we need to check
    LOGGER.info("Waiting for DCM pitch adjust to complete...")

    # DCM Roll
    dcm_roll_adjuster = lookup_table_adjuster(
        linear_interpolation_lut(
            *parse_lookup_table(undulator_dcm.roll_energy_table_path)
        ),
        dcm.roll_in_mrad,
        bragg_deg,
    )
    yield from dcm_roll_adjuster(DCM_GROUP)
    LOGGER.info("Waiting for DCM roll adjust to complete...")

    #
    # Adjust vfm mirror stripe and mirror voltages
    #

    # VFM Stripe selection
    yield from adjust_mirror_stripe(energy_kev, vfm, mirror_voltages)



from importlib import resources
from pathlib import Path
from unittest.mock import patch

import pytest

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from tests.conftest import raw_params_from_file

BANNED_PATHS = [Path("/dls"), Path("/dls_sw")]


@pytest.fixture
def load_centre_collect_params():
    json_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json"
    )
    return LoadCentreCollect(**json_dict)


@pytest.fixture(autouse=True)
def patch_open_to_prevent_dls_reads_in_tests():
    unpatched_open = open
    assert __package__
    project_folder = resources.files(__package__)
    assert isinstance(project_folder, Path)
    project_folder = project_folder.parent.parent.parent

    def patched_open(*args, **kwargs):
        requested_path = Path(args[0])
        if requested_path.is_absolute():
            for p in BANNED_PATHS:
                assert not requested_path.is_relative_to(
                    p
                ) or requested_path.is_relative_to(project_folder), (
                    f"Attempt to open {requested_path} from inside a unit test"
                )
        return unpatched_open(*args, **kwargs)

    with patch("builtins.open", side_effect=patched_open):
        yield []


@pytest.fixture
def dummy_rotation_data_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="SAD",
        sample_id=364758,
    )



import os

from dodal.devices.detector import EIGER2_X_16M_SIZE
from pydantic.dataclasses import dataclass

from mx_bluesky.common.parameters.constants import (
    DeviceSettingsConstants,
    DocDescriptorNames,
    EnvironmentConstants,
    ExperimentParamConstants,
    HardwareConstants,
    OavConstants,
    PlanGroupCheckpointConstants,
    PlanNameConstants,
    SimConstants,
    TriggerConstants,
)

TEST_MODE = os.environ.get("HYPERION_TEST_MODE")


@dataclass(frozen=True)
class I03Constants:
    BASE_DATA_DIR = "/tmp/dls/i03/data/" if TEST_MODE else "/dls/i03/data/"
    BEAMLINE = "BL03S" if TEST_MODE else "BL03I"
    DETECTOR = EIGER2_X_16M_SIZE
    INSERTION_PREFIX = "SR03S" if TEST_MODE else "SR03I"
    OAV_CENTRING_FILE = OavConstants.OAV_CONFIG_JSON
    SHUTTER_TIME_S = 0.06
    USE_PANDA_FOR_GRIDSCAN = False
    SET_STUB_OFFSETS = False
    OMEGA_FLIP = True

    # Turns on GPU processing for zocalo and logs a comparison between GPU and CPU-
    # processed results.
    COMPARE_CPU_AND_GPU_ZOCALO = False

    # Turns on GPU processing for zocalo and uses the results that come back
    USE_GPU_RESULTS = False


@dataclass(frozen=True)
class HyperionConstants:
    DESCRIPTORS = DocDescriptorNames()
    TRIGGER = TriggerConstants()
    ZOCALO_ENV = EnvironmentConstants.ZOCALO_ENV
    HARDWARE = HardwareConstants()
    I03 = I03Constants()
    PARAM = ExperimentParamConstants()
    PLAN = PlanNameConstants()
    WAIT = PlanGroupCheckpointConstants()
    SIM = SimConstants()
    TRIGGER = TriggerConstants()
    CALLBACK_0MQ_PROXY_PORTS = (5577, 5578)
    DESCRIPTORS = DocDescriptorNames()
    CONFIG_SERVER_URL = (
        "http://fake-url-not-real"
        if TEST_MODE
        else "https://daq-config.diamond.ac.uk/api"
    )
    GRAYLOG_PORT = 12232
    PARAMETER_SCHEMA_DIRECTORY = "src/hyperion/parameters/schemas/"
    LOG_FILE_NAME = "hyperion.log"
    DEVICE_SETTINGS_CONSTANTS = DeviceSettingsConstants()


CONST = HyperionConstants()



from unittest.mock import ANY, MagicMock, call, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import assert_message_and_return_remaining
from bluesky.utils import IllegalMessageSequence

from mx_bluesky.common.parameters.constants import (
    PlanNameConstants,
)
from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_decorator,
)
from tests.conftest import RunEngineSimulator, XBPMAndTransmissionWrapperComposite


def assert_open_run_then_pause_xbpm_then_close_run_then_unpause(msgs):
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "trigger" and msg.obj.name == "xbpm_feedback",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "close_run"
        and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args == (1.0,),
    )


def test_xbpm_preprocessor_does_nothing_on_non_specified_message(
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    sim_run_engine: RunEngineSimulator,
):
    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite,
        desired_transmission_fraction=1,
        run_key_to_wrap=PlanNameConstants.GRIDSCAN_OUTER,
    )
    @bpp.set_run_key_decorator(PlanNameConstants.DO_FGS)
    @bpp.run_decorator()
    def my_boring_plan():
        yield from bps.null()

    msgs = sim_run_engine.simulate_plan(my_boring_plan())

    assert len(msgs) == 3
    assert msgs[0].command == "open_run"
    assert msgs[1].command == "null"
    assert msgs[2].command == "close_run"


def test_xbpm_preprocessor_does_nothing_with_no_run(
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    sim_run_engine: RunEngineSimulator,
):
    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite,
        desired_transmission_fraction=1,
    )
    def my_boring_plan():
        yield from bps.null()

    msgs = sim_run_engine.simulate_plan(my_boring_plan())

    assert len(msgs) == 1
    assert msgs[0].command == "null"


def test_xbpm_preprocessor_runs_inserts_correct_plan_on_correct_message(
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    sim_run_engine: RunEngineSimulator,
):
    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite,
        desired_transmission_fraction=1,
        run_key_to_wrap=PlanNameConstants.GRIDSCAN_OUTER,
    )
    @bpp.set_run_key_decorator(PlanNameConstants.GRIDSCAN_OUTER)
    @bpp.run_decorator()
    def open_run_plan():
        yield from bps.null()

    msgs = sim_run_engine.simulate_plan(open_run_plan())
    assert_open_run_then_pause_xbpm_then_close_run_then_unpause(msgs)


@patch(
    "mx_bluesky.common.preprocessors.preprocessors.unpause_xbpm_feedback_and_set_transmission_to_1"
)
def test_xbpm_preprocessor_unpauses_xbpm_on_exception(
    mock_unpause_xbpm: MagicMock,
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    sim_run_engine: RunEngineSimulator,
):
    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite, desired_transmission_fraction=1
    )
    @bpp.set_run_key_decorator(PlanNameConstants.GRIDSCAN_OUTER)
    @bpp.run_decorator()
    def except_plan():
        yield from bps.null()
        raise Exception

    with pytest.raises(Exception):  # noqa: B017
        sim_run_engine.simulate_plan(except_plan())

    # Called once on exception and once on close_run
    mock_unpause_xbpm.assert_has_calls([call(ANY, ANY)])


@patch("mx_bluesky.common.preprocessors.preprocessors.check_and_pause_feedback")
@patch(
    "mx_bluesky.common.preprocessors.preprocessors.unpause_xbpm_feedback_and_set_transmission_to_1"
)
def test_xbpm_preprocessor_wraps_one_run_only_if_no_run_specified(
    mock_unpause_xbpm: MagicMock,
    mock_pause: MagicMock,
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    RE: RunEngine,
):
    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite, desired_transmission_fraction=1
    )
    @bpp.run_decorator()
    def first_plan():
        mock_pause.assert_called_once()
        yield from second_plan()

        # Check we didn't unpause on the inner-run
        mock_unpause_xbpm.assert_not_called()

    @bpp.set_run_key_decorator(PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN)
    @bpp.run_decorator()
    def second_plan():
        yield from bps.null()

    RE(first_plan())
    mock_pause.assert_called_once()
    mock_unpause_xbpm.assert_called_once()


def test_xbpm_preprocessor_cant_unpause_on_wrong_run(
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    RE: RunEngine,
):
    # Logic in the preprocessor relies on the assumption that Bluesky doesn't let us have
    # multiple unnamed runs open, or multiple runs with the same name

    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite, desired_transmission_fraction=1
    )
    @bpp.run_decorator()
    def first_unnamed_run():
        yield from second_unnamed_run()

    @bpp.run_decorator()
    def second_unnamed_run():
        yield from bps.null()

    @bpp.set_run_key_decorator(PlanNameConstants.GRID_DETECT_INNER)
    @bpp.run_decorator()
    @transmission_and_xbpm_feedback_for_collection_decorator(
        devices=xbpm_and_transmission_wrapper_composite,
        desired_transmission_fraction=1,
    )
    def first_named_run():
        yield from second_named_run()

    @bpp.set_run_key_decorator(PlanNameConstants.GRID_DETECT_INNER)
    @bpp.run_decorator()
    def second_named_run():
        yield from bps.null()

    with pytest.raises(IllegalMessageSequence):
        RE(first_unnamed_run())

    with pytest.raises(IllegalMessageSequence):
        RE(first_named_run())



from __future__ import annotations

import bluesky.plan_stubs as bps
import numpy as np
import pydantic
from blueapi.core import BlueskyContext
from bluesky.preprocessors import run_decorator, set_run_key_decorator, subs_wrapper
from bluesky.utils import MsgGenerator
from dodal.devices.oav.oav_parameters import OAVParameters

import mx_bluesky.common.xrc_result as flyscan_result
from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.xrc_result import XRayCentreEventHandler
from mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan import (
    RobotLoadThenCentreComposite,
    robot_load_then_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    MultiRotationScan,
    RotationScanComposite,
    multi_rotation_scan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class LoadCentreCollectComposite(RobotLoadThenCentreComposite, RotationScanComposite):
    """Composite that provides access to the required devices."""


def create_devices(context: BlueskyContext) -> LoadCentreCollectComposite:
    """Create the necessary devices for the plan."""
    return device_composite_from_context(context, LoadCentreCollectComposite)


def load_centre_collect_full(
    composite: LoadCentreCollectComposite,
    parameters: LoadCentreCollect,
    oav_params: OAVParameters | None = None,
) -> MsgGenerator:
    """Attempt a complete data collection experiment, consisting of the following:
    * Load the sample if necessary
    * Move to the specified goniometer start angles
    * Perform optical centring, then X-ray centring
    * If X-ray centring finds a diffracting centre then move to that centre and
    * do a collection with the specified parameters.
    """
    if not oav_params:
        oav_params = OAVParameters(context="xrayCentring")

    @set_run_key_decorator(CONST.PLAN.LOAD_CENTRE_COLLECT)
    @run_decorator(
        md={
            "metadata": {"sample_id": parameters.sample_id},
            "activate_callbacks": ["BeamDrawingCallback", "SampleHandlingCallback"],
            "with_snapshot": parameters.multi_rotation_scan.model_dump_json(
                include=WithSnapshot.model_fields.keys()  # type: ignore
            ),
        }
    )
    def plan_with_callback_subs():
        flyscan_event_handler = XRayCentreEventHandler()
        yield from subs_wrapper(
            robot_load_then_xray_centre(composite, parameters.robot_load_then_centre),
            flyscan_event_handler,
        )

        locations_to_collect_um: list[np.ndarray] = []

        if flyscan_event_handler.xray_centre_results:
            selection_func = flyscan_result.resolve_selection_fn(
                parameters.selection_params
            )
            hits = selection_func(flyscan_event_handler.xray_centre_results)
            locations_to_collect_um = [hit.centre_of_mass_mm * 1000 for hit in hits]

            LOGGER.info(
                f"Selected hits {hits} using {selection_func}, args={parameters.selection_params}"
            )
        else:
            # If the xray centring hasn't found a result but has not thrown an error it
            # means that we do not need to recentre and can collect where we are
            initial_x = yield from bps.rd(composite.smargon.x.user_readback)
            initial_y = yield from bps.rd(composite.smargon.y.user_readback)
            initial_z = yield from bps.rd(composite.smargon.z.user_readback)

            locations_to_collect_um = [np.array([initial_x, initial_y, initial_z])]

        multi_rotation = parameters.multi_rotation_scan
        rotation_template = multi_rotation.rotation_scans.copy()

        multi_rotation.rotation_scans.clear()

        for location in locations_to_collect_um:
            for rot in rotation_template:
                combination = rot.model_copy()
                (
                    combination.x_start_um,
                    combination.y_start_um,
                    combination.z_start_um,
                ) = location
                multi_rotation.rotation_scans.append(combination)
        multi_rotation = MultiRotationScan.model_validate(multi_rotation)

        assert (
            multi_rotation.demand_energy_ev
            == parameters.robot_load_then_centre.demand_energy_ev
        ), "Setting a different energy for gridscan and rotation is not supported"
        yield from multi_rotation_scan(composite, multi_rotation, oav_params)

    yield from plan_with_callback_subs()



from unittest.mock import MagicMock, call, patch

import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue

from mx_bluesky.hyperion.device_setup_plans.manipulate_sample import (
    move_aperture_if_required,
    move_phi_chi_omega,
    move_x_y_z,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan


@pytest.mark.parametrize(
    "set_position",
    [
        (ApertureValue.SMALL),
        (ApertureValue.MEDIUM),
        (ApertureValue.OUT_OF_BEAM),
        (ApertureValue.LARGE),
    ],
)
async def test_move_aperture_goes_to_correct_position(
    aperture_scatterguard: ApertureScatterguard,
    RE: RunEngine,
    set_position,
):
    with patch.object(aperture_scatterguard, "set") as mock_set:
        RE(move_aperture_if_required(aperture_scatterguard, set_position))
        mock_set.assert_called_once_with(
            set_position,
        )


async def test_move_aperture_does_nothing_when_none_selected(
    aperture_scatterguard: ApertureScatterguard, RE: RunEngine
):
    with patch.object(aperture_scatterguard, "set") as mock_set:
        RE(move_aperture_if_required(aperture_scatterguard, None))
        mock_set.assert_not_called()


@pytest.mark.parametrize(
    "motor_position, expected_moves",
    [
        [[1, 2, 3], [1, 2, 3]],
        [[0, 0, 0], [None, None, None]],
        [[None, None, None], [None, None, None]],
        [[1, 0, 0], [1, 0, 0]],
        [[0, 1, 0], [0, 1, 0]],
        [[0, 0, 1], [0, 0, 1]],
        [[1, None, None], [1, None, None]],
        [[None, 1, None], [None, 1, None]],
        [[None, None, 1], [None, None, 1]],
    ],
)
@patch("bluesky.plan_stubs.abs_set", autospec=True)
def test_move_x_y_z(
    bps_abs_set: MagicMock,
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
    fake_fgs_composite: HyperionFlyScanXRayCentreComposite,
    RE: RunEngine,
    motor_position: list[float],
    expected_moves: list[float | None],
):
    RE(move_x_y_z(fake_fgs_composite.smargon, *motor_position))  # type: ignore
    expected_calls = [
        call(axis, pos, group="move_x_y_z")
        for axis, pos in zip(
            [
                fake_fgs_composite.smargon.x,
                fake_fgs_composite.smargon.y,
                fake_fgs_composite.smargon.z,
            ],
            expected_moves,
            strict=False,
        )
        if pos is not None
    ]
    bps_abs_set.assert_has_calls(
        expected_calls,
        any_order=True,
    )


@pytest.mark.parametrize(
    "motor_position, expected_moves",
    [
        [[1, 2, 3], [1, 2, 3]],
        [[0, 0, 0], [0, 0, 0]],
        [[0, None, None], [0, None, None]],
        [[None, 0, None], [None, 0, None]],
        [[None, None, 0], [None, None, 0]],
        [[None, None, None], [None, None, None]],
        [[1, 0, 0], [1, 0, 0]],
    ],
)
@patch("bluesky.plan_stubs.abs_set", autospec=True)
def test_move_phi_chi_omega(
    bps_abs_set: MagicMock,
    test_fgs_params: HyperionSpecifiedThreeDGridScan,
    fake_fgs_composite: HyperionFlyScanXRayCentreComposite,
    RE: RunEngine,
    motor_position: list[float],
    expected_moves: list[float | None],
):
    RE(move_phi_chi_omega(fake_fgs_composite.smargon, *motor_position))  # type: ignore
    expected_calls = [
        call(axis, pos, group="move_phi_chi_omega")
        for axis, pos in zip(
            [
                fake_fgs_composite.smargon.phi,
                fake_fgs_composite.smargon.chi,
                fake_fgs_composite.smargon.omega,
            ],
            expected_moves,
            strict=False,
        )
        if pos is not None
    ]
    bps_abs_set.assert_has_calls(
        expected_calls,
        any_order=True,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from mx_bluesky.common.external_interaction.callbacks.common.logging_callback import (
    format_doc_for_log,
)
from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.external_interaction.nexus.nexus_utils import (
    AxisDirection,
    create_beam_and_attenuator_parameters,
    vds_type_based_on_bit_depth,
)
from mx_bluesky.common.external_interaction.nexus.write_nexus import NexusWriter
from mx_bluesky.common.utils.log import NEXUS_LOGGER
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart


class RotationNexusFileCallback(PlanReactiveCallback):
    """Callback class to handle the creation of Nexus files based on experiment
    parameters for rotation scans

    To use, subscribe the Bluesky RunEngine to an instance of this class.
    E.g.:
        nexus_file_handler_callback = NexusFileCallback(parameters)
        RE.subscribe(nexus_file_handler_callback)
    Or decorate a plan using bluesky.preprocessors.subs_decorator.

    See: https://blueskyproject.io/bluesky/callbacks.html#ways-to-invoke-callbacks
    """

    def __init__(self) -> None:
        super().__init__(NEXUS_LOGGER)
        self.run_uid: str | None = None
        self.writer: NexusWriter | None = None
        self.descriptors: dict[str, EventDescriptor] = {}
        # used when multiple collections are made in one detector arming event:
        self.full_num_of_images: int | None = None
        self.meta_data_run_number: int | None = None

    def activity_gated_descriptor(self, doc: EventDescriptor):
        self.descriptors[doc["uid"]] = doc

    def activity_gated_event(self, doc: Event):
        event_descriptor = self.descriptors.get(doc["descriptor"])
        if event_descriptor is None:
            NEXUS_LOGGER.warning(
                f"Rotation Nexus handler {self} received event doc {format_doc_for_log(doc)} and "
                "has no corresponding descriptor record"
            )
            return doc
        if event_descriptor.get("name") == CONST.DESCRIPTORS.HARDWARE_READ_DURING:
            NEXUS_LOGGER.info(
                f"Nexus handler received event from read hardware {format_doc_for_log(doc)}"
            )
            data = doc["data"]
            assert self.writer, "Nexus writer not initialised"
            (
                self.writer.beam,
                self.writer.attenuator,
            ) = create_beam_and_attenuator_parameters(
                data["dcm-energy_in_kev"],
                data["flux-flux_reading"],
                data["attenuator-actual_transmission"],
            )
            vds_data_type = vds_type_based_on_bit_depth(doc["data"]["eiger_bit_depth"])
            self.writer.create_nexus_file(vds_data_type)
            NEXUS_LOGGER.info(f"Nexus file created at {self.writer.data_filename}")
        return doc

    def activity_gated_start(self, doc: RunStart):
        if doc.get("subplan_name") == CONST.PLAN.ROTATION_MULTI:
            self.full_num_of_images = doc.get("full_num_of_images")
            self.meta_data_run_number = doc.get("meta_data_run_number")
        if doc.get("subplan_name") == CONST.PLAN.ROTATION_OUTER:
            self.run_uid = doc.get("uid")
            hyperion_params = doc.get("mx_bluesky_parameters")
            assert isinstance(hyperion_params, str)
            NEXUS_LOGGER.info(
                f"Nexus writer received start document with experiment parameters {hyperion_params}"
            )
            parameters = RotationScan.model_validate_json(hyperion_params)
            NEXUS_LOGGER.info("Setting up nexus file...")

            det_size = (
                parameters.detector_params.detector_size_constants.det_size_pixels
            )
            shape = (parameters.num_images, det_size.width, det_size.height)
            self.writer = NexusWriter(
                parameters,
                shape,
                parameters.scan_points,
                omega_start_deg=parameters.omega_start_deg,
                chi_start_deg=parameters.chi_start_deg or 0,
                phi_start_deg=parameters.phi_start_deg or 0,
                vds_start_index=parameters.nexus_vds_start_img,
                full_num_of_images=self.full_num_of_images,
                meta_data_run_number=self.meta_data_run_number,
                axis_direction=AxisDirection.NEGATIVE
                if parameters.features.omega_flip
                else AxisDirection.POSITIVE,
            )



from copy import deepcopy

import pytest

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionPositionInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import StoreInIspyb
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

from ......conftest import (
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
    TEST_SAMPLE_ID,
    default_raw_gridscan_params,
)


@pytest.fixture
def dummy_params():
    dummy_params = HyperionSpecifiedThreeDGridScan(**default_raw_gridscan_params())
    dummy_params.sample_id = TEST_SAMPLE_ID
    dummy_params.run_number = 0
    return dummy_params


@pytest.fixture
def dummy_3d_gridscan_ispyb():
    store_in_ispyb_3d = StoreInIspyb(CONST.SIM.ISPYB_CONFIG)
    return store_in_ispyb_3d


@pytest.fixture
def dummy_rotation_ispyb(dummy_rotation_params):
    store_in_ispyb = StoreInIspyb(CONST.SIM.ISPYB_CONFIG)
    return store_in_ispyb


@pytest.fixture
def dummy_2d_gridscan_ispyb():
    return StoreInIspyb(CONST.SIM.ISPYB_CONFIG)


@pytest.fixture
def scan_xy_data_info_for_update(
    scan_data_info_for_begin: ScanDataInfo,
) -> ScanDataInfo:
    scan_data_info_for_update = deepcopy(scan_data_info_for_begin)
    scan_data_info_for_update.data_collection_info.parent_id = (
        TEST_DATA_COLLECTION_GROUP_ID
    )
    scan_data_info_for_update.data_collection_info.synchrotron_mode = "test"
    scan_data_info_for_update.data_collection_info.flux = 10
    scan_data_info_for_update.data_collection_grid_info = DataCollectionGridInfo(
        dx_in_mm=0.1,
        dy_in_mm=0.1,
        steps_x=40,
        steps_y=20,
        microns_per_pixel_x=1.25,
        microns_per_pixel_y=1.25,
        snapshot_offset_x_pixel=100,
        snapshot_offset_y_pixel=100,
        orientation=Orientation.HORIZONTAL,
        snaked=True,
    )
    scan_data_info_for_update.data_collection_position_info = (
        DataCollectionPositionInfo(0, 0, 0)
    )
    scan_data_info_for_update.data_collection_id = TEST_DATA_COLLECTION_IDS[0]
    return scan_data_info_for_update



import dataclasses
from collections.abc import Sequence
from typing import Any
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from bluesky.protocols import Location
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.i03.beamstop import BeamstopPositions
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.synchrotron import SynchrotronMode
from ophyd.sim import NullStatus
from ophyd_async.testing import set_mock_value
from pydantic import ValidationError

from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.hyperion.device_setup_plans.check_beamstop import BeamstopException
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    CrystalNotFoundException,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
    load_centre_collect_full,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan import (
    RobotLoadThenCentreComposite,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadAndEnergyChange
from mx_bluesky.hyperion.parameters.rotation import (
    MultiRotationScan,
    RotationScanPerSweep,
)

from ....conftest import pin_tip_edge_data, raw_params_from_file
from .conftest import (
    FLYSCAN_RESULT_HIGH,
    FLYSCAN_RESULT_LOW,
    FLYSCAN_RESULT_MED,
    sim_fire_event_on_open_run,
)

GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION = "tests/test_data/parameter_json_files/good_test_load_centre_collect_params_multi_rotation.json"


@pytest.fixture
def composite(
    robot_load_composite,
    fake_create_rotation_devices,
    pin_tip_detection_with_found_pin,
    sim_run_engine: RunEngineSimulator,
) -> LoadCentreCollectComposite:
    rlaec_args = {
        field.name: getattr(robot_load_composite, field.name)
        for field in dataclasses.fields(robot_load_composite)
    }
    rotation_args = {
        field.name: getattr(fake_create_rotation_devices, field.name)
        for field in dataclasses.fields(fake_create_rotation_devices)
    }

    composite = LoadCentreCollectComposite(**(rlaec_args | rotation_args))
    composite.pin_tip_detection = pin_tip_detection_with_found_pin
    composite.undulator_dcm.set = MagicMock(return_value=NullStatus())
    minaxis = Location(setpoint=-2, readback=-2)
    maxaxis = Location(setpoint=2, readback=2)
    tip_x_px, tip_y_px, top_edge_array, bottom_edge_array = pin_tip_edge_data()
    sim_run_engine.add_handler(
        "locate", lambda _: minaxis, "smargon-x-low_limit_travel"
    )
    sim_run_engine.add_handler(
        "locate", lambda _: minaxis, "smargon-y-low_limit_travel"
    )
    sim_run_engine.add_handler(
        "locate", lambda _: minaxis, "smargon-z-low_limit_travel"
    )
    sim_run_engine.add_handler(
        "locate", lambda _: maxaxis, "smargon-x-high_limit_travel"
    )
    sim_run_engine.add_handler(
        "locate", lambda _: maxaxis, "smargon-y-high_limit_travel"
    )
    sim_run_engine.add_handler(
        "locate", lambda _: maxaxis, "smargon-z-high_limit_travel"
    )
    sim_run_engine.add_read_handler_for(
        composite.synchrotron.synchrotron_mode, SynchrotronMode.USER
    )
    sim_run_engine.add_read_handler_for(
        composite.synchrotron.top_up_start_countdown, -1
    )
    sim_run_engine.add_read_handler_for(
        composite.pin_tip_detection.triggered_top_edge, top_edge_array
    )
    sim_run_engine.add_read_handler_for(
        composite.pin_tip_detection.triggered_bottom_edge, bottom_edge_array
    )
    zoom_levels_list = ["1.0x", "3.0x", "5.0x", "7.5x", "10.0x"]
    composite.oav.zoom_controller.level.describe = AsyncMock(
        return_value={"level": {"choices": zoom_levels_list}}
    )
    set_mock_value(composite.oav.zoom_controller.level, "7.5x")

    sim_run_engine.add_read_handler_for(
        composite.pin_tip_detection.triggered_tip, (tip_x_px, tip_y_px)
    )
    return composite


@pytest.fixture
def load_centre_collect_params_multi():
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    return LoadCentreCollect(**params)


@pytest.fixture
def load_centre_collect_with_top_n_params():
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/load_centre_collect_params_top_n_by_max_count.json"
    )
    return LoadCentreCollect(**params)


def test_can_serialize_load_centre_collect_params(load_centre_collect_params):
    load_centre_collect_params.model_dump_json()


def test_params_good_multi_rotation_load_centre_collect_params(
    load_centre_collect_params_multi,
):
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    LoadCentreCollect(**params)


def test_params_with_varying_frames_per_rotation_is_rejected():
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    params["multi_rotation_scan"]["rotation_scans"][0]["scan_width_deg"] = 180
    params["multi_rotation_scan"]["rotation_scans"][1]["scan_width_deg"] = 90
    with pytest.raises(
        ValidationError,
        match="Sweeps with different numbers of frames are not supported.",
    ):
        LoadCentreCollect(**params)


@pytest.mark.parametrize(
    "param, value",
    [
        ["x_start_um", 1.0],
        ["y_start_um", 2.0],
        ["z_start_um", 3.0],
    ],
)
def test_params_with_start_xyz_is_rejected(param: str, value: float):
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    params["multi_rotation_scan"]["rotation_scans"][1][param] = value
    with pytest.raises(
        ValidationError,
        match="Specifying start xyz for sweeps is not supported in combination with centring.",
    ):
        LoadCentreCollect(**params)


def test_params_with_different_energy_for_rotation_gridscan_rejected():
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    params["multi_rotation_scan"]["demand_energy_ev"] = 11000
    params["robot_load_then_centre"]["demand_energy_ev"] = 11100
    with pytest.raises(
        ValidationError,
        match="Setting a different energy for gridscan and rotation is not supported.",
    ):
        LoadCentreCollect(**params)


@pytest.mark.parametrize(
    "key, value",
    [
        # MxBlueskyParameters
        ["parameter_model_version", "1.2.3"],
        # WithSample
        ["sample_id", 12345],
        ["sample_puck", 1],
        ["sample_pin", 2],
        # WithVisit
        ["beamline", "i03"],
        ["visit", "cm12345"],
        ["insertion_prefix", "SR03"],
        ["detector_distance_mm", 123],
        ["det_dist_to_beam_converter_path", "/foo/bar"],
    ],
)
def test_params_with_unexpected_info_in_robot_load_rejected(key: str, value: Any):
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    params["robot_load_then_centre"][key] = value
    with pytest.raises(
        ValidationError, match="Unexpected keys in robot_load_then_centre"
    ):
        LoadCentreCollect(**params)


@pytest.mark.parametrize(
    "key, value",
    [
        # MxBlueskyParameters
        ["parameter_model_version", "1.2.3"],
        # WithSample
        ["sample_id", 12345],
        ["sample_puck", 1],
        ["sample_pin", 2],
        # WithVisit
        ["beamline", "i03"],
        ["visit", "cm12345"],
        ["insertion_prefix", "SR03"],
        ["detector_distance_mm", 123],
        ["det_dist_to_beam_converter_path", "/foo/bar"],
    ],
)
def test_params_with_unexpected_info_in_multi_rotation_scan_rejected(
    key: str, value: Any
):
    params = raw_params_from_file(GOOD_TEST_LOAD_CENTRE_COLLECT_MULTI_ROTATION)
    params["multi_rotation_scan"][key] = value
    with pytest.raises(ValidationError, match="Unexpected keys in multi_rotation_scan"):
        LoadCentreCollect(**params)


def test_can_serialize_load_centre_collect_robot_load_params(
    load_centre_collect_params,
):
    load_centre_collect_params.robot_load_then_centre.model_dump_json()


def test_can_serialize_load_centre_collect_multi_rotation_scan(
    load_centre_collect_params,
):
    load_centre_collect_params.multi_rotation_scan.model_dump_json()


def test_can_serialize_load_centre_collect_single_rotation_scans(
    load_centre_collect_params,
):
    list(load_centre_collect_params.multi_rotation_scan.single_rotation_scans)[
        0
    ].model_dump_json()


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    return_value=iter(
        [
            Msg(
                "open_run",
                xray_centre_results=[dataclasses.asdict(FLYSCAN_RESULT_MED)],
                run=CONST.PLAN.FLYSCAN_RESULTS,
            ),
            Msg("close_run"),
        ]
    ),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    return_value=iter([Msg(command="robot_load_and_change_energy")]),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan.multi_rotation_scan",
    return_value=iter([Msg(command="multi_rotation_scan")]),
)
def test_collect_full_plan_happy_path_invokes_all_steps_and_centres_on_best_flyscan_result(
    mock_rotation_scan: MagicMock,
    mock_full_robot_load_plan: MagicMock,
    mock_pin_centre_then_xray_centre_plan: MagicMock,
    composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    msgs = sim_run_engine.simulate_plan(
        load_centre_collect_full(
            composite, load_centre_collect_params, oav_parameters_for_rotation
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run" and "xray_centre_results" in msg.kwargs,
    )
    # TODO re-enable tests see mx-bluesky 561
    # msgs = assert_message_and_return_remaining(
    #     msgs, lambda msg: msg.command == "set" and msg.args[0] == ApertureValue.MEDIUM
    # )
    # msgs = assert_message_and_return_remaining(
    #     msgs,
    #     lambda msg: msg.command == "set"
    #     and msg.obj.name == "smargon-x"
    #     and msg.args[0] == 0.1,
    # )
    # msgs = assert_message_and_return_remaining(
    #     msgs,
    #     lambda msg: msg.command == "set"
    #     and msg.obj.name == "smargon-y"
    #     and msg.args[0] == 0.2,
    # )
    # msgs = assert_message_and_return_remaining(
    #     msgs,
    #     lambda msg: msg.command == "set"
    #     and msg.obj.name == "smargon-z"
    #     and msg.args[0] == 0.3,
    # )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "multi_rotation_scan"
    )

    robot_load_energy_change_composite = mock_full_robot_load_plan.mock_calls[0].args[0]
    robot_load_energy_change_params = mock_full_robot_load_plan.mock_calls[0].args[1]
    assert isinstance(robot_load_energy_change_composite, RobotLoadThenCentreComposite)
    assert isinstance(robot_load_energy_change_params, RobotLoadAndEnergyChange)
    mock_pin_centre_then_xray_centre_plan.assert_called_once()
    mock_rotation_scan.assert_called_once()
    rotation_scan_composite = mock_rotation_scan.mock_calls[0].args[0]
    rotation_scan_params = mock_rotation_scan.mock_calls[0].args[1]
    assert isinstance(rotation_scan_composite, RotationScanComposite)
    assert isinstance(rotation_scan_params, MultiRotationScan)
    # XXX sample test file xyz conflicts with detected xyz
    # see https://github.com/DiamondLightSource/mx-bluesky/issues/563
    expected_rotation_scans = [
        {
            "omega_start_deg": 0,
            "chi_start_deg": 23.85,
            "x_start_um": 400,
            "y_start_um": 500,
            "z_start_um": 600,
            "nexus_vds_start_img": 0,
        },
    ]
    _compare_rotation_scans(
        expected_rotation_scans, rotation_scan_params.rotation_scans
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan.multi_rotation_scan",
    return_value=iter([]),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    new=MagicMock(),
)
def test_load_centre_collect_full_skips_collect_if_pin_tip_not_found(
    mock_rotation_scan: MagicMock,
    composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_read_handler_for(
        composite.pin_tip_detection.triggered_tip, PinTipDetection.INVALID_POSITION
    )
    sim_run_engine.add_read_handler_for(composite.oav.microns_per_pixel_x, 1.58)
    sim_run_engine.add_read_handler_for(composite.oav.microns_per_pixel_y, 1.58)

    with pytest.raises(WarningException, match="Pin tip centring failed"):
        sim_run_engine.simulate_plan(
            load_centre_collect_full(
                composite, load_centre_collect_params, oav_parameters_for_rotation
            )
        )

    mock_rotation_scan.assert_not_called()


@patch(
    "mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan.multi_rotation_scan",
    return_value=iter([]),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    new=MagicMock(),
)
def test_load_centre_collect_full_plan_skips_collect_if_no_diffraction(
    mock_rotation_scan: MagicMock,
    composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    sim_run_engine: RunEngineSimulator,
    grid_detection_callback_with_detected_grid,
):
    sim_run_engine.add_read_handler_for(composite.oav.microns_per_pixel_x, 1.58)
    sim_run_engine.add_read_handler_for(composite.oav.microns_per_pixel_y, 1.58)

    with pytest.raises(CrystalNotFoundException):
        sim_run_engine.simulate_plan(
            load_centre_collect_full(
                composite, load_centre_collect_params, oav_parameters_for_rotation
            )
        )

    mock_rotation_scan.assert_not_called()


def test_load_centre_collect_fails_with_exception_when_no_beamstop(
    composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_read_handler_for(
        composite.beamstop.selected_pos, BeamstopPositions.UNKNOWN
    )

    with pytest.raises(BeamstopException):
        sim_run_engine.simulate_plan(
            load_centre_collect_full(
                composite, load_centre_collect_params, oav_parameters_for_rotation
            )
        )


def test_can_deserialize_top_n_by_max_count_params(
    load_centre_collect_with_top_n_params,
):
    assert load_centre_collect_with_top_n_params.select_centres.name == "TopNByMaxCount"
    assert load_centre_collect_with_top_n_params.select_centres.n == 5


def test_bad_selection_method_is_rejected():
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/load_centre_collect_params_top_n_by_max_count.json"
    )
    params["select_centres"]["name"] = "inject_bad_code_here"
    with pytest.raises(
        ValidationError,
        match=(
            "Input tag 'inject_bad_code_here' found using 'name' does not match any "
            "of the expected tags"
        ),
    ):
        LoadCentreCollect(**params)


def test_default_select_centres_is_top_n_by_max_count_n_is_1(
    load_centre_collect_params,
):
    assert load_centre_collect_params.select_centres.name == "TopNByMaxCount"
    assert load_centre_collect_params.select_centres.n == 1


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.pin_centre_then_flyscan_plan",
    new=MagicMock(
        return_value=iter(
            [
                Msg(
                    "open_run",
                    xray_centre_results=[
                        dataclasses.asdict(r)
                        for r in [
                            FLYSCAN_RESULT_MED,
                            FLYSCAN_RESULT_HIGH,
                            FLYSCAN_RESULT_MED,
                            FLYSCAN_RESULT_LOW,
                            FLYSCAN_RESULT_MED,
                            FLYSCAN_RESULT_HIGH,
                        ]
                    ],
                    run=CONST.PLAN.FLYSCAN_RESULTS,
                ),
                Msg("close_run"),
            ]
        )
    ),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan.robot_load_and_change_energy_plan",
    new=MagicMock(return_value=iter([Msg(command="robot_load_and_change_energy")])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan.multi_rotation_scan",
    side_effect=lambda _, __, ___: iter([Msg(command="multi_rotation_scan")]),
)
def test_load_centre_collect_full_plan_multiple_centres(
    mock_multi_rotation_scan: MagicMock,
    sim_run_engine: RunEngineSimulator,
    load_centre_collect_with_top_n_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    composite: LoadCentreCollectComposite,
):
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, CONST.PLAN.FLYSCAN_RESULTS)
    msgs = sim_run_engine.simulate_plan(
        load_centre_collect_full(
            composite,
            load_centre_collect_with_top_n_params,
            oav_parameters_for_rotation,
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "robot_load_and_change_energy"
    )
    assert sum(1 for msg in msgs if msg.command == "robot_load_and_change_energy") == 1
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run" and "xray_centre_results" in msg.kwargs,
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "multi_rotation_scan"
    )

    def _rotation_at_first_position(chi=0):
        return {
            "omega_start_deg": 10,
            "chi_start_deg": chi,
            "x_start_um": 100,
            "y_start_um": 200,
            "z_start_um": 300,
        }

    def _rotation_at_second_position(chi=0):
        return {
            "omega_start_deg": 10,
            "chi_start_deg": chi,
            "x_start_um": 400,
            "y_start_um": 500,
            "z_start_um": 600,
        }

    expected_rotation_scans = [
        _rotation_at_first_position(0),
        _rotation_at_first_position(30),
        _rotation_at_first_position(0),
        _rotation_at_first_position(30),
        _rotation_at_second_position(0),
        _rotation_at_second_position(30),
        _rotation_at_second_position(0),
        _rotation_at_second_position(30),
        _rotation_at_second_position(0),
        _rotation_at_second_position(30),
    ]
    for i in range(0, len(expected_rotation_scans)):
        expected_rotation_scans[i]["nexus_vds_start_img"] = 3600 * i

    rotation_scan_params = mock_multi_rotation_scan.mock_calls[0].args[1]
    assert isinstance(rotation_scan_params, MultiRotationScan)
    _compare_rotation_scans(
        expected_rotation_scans, rotation_scan_params.rotation_scans
    )
    assert rotation_scan_params.transmission_frac == 0.05


def _compare_rotation_scans(
    expected_rotation_scans: Sequence[dict],
    actual_rotation_scans: Sequence[RotationScanPerSweep],
):
    for expected, rotation_scan in zip(
        expected_rotation_scans, actual_rotation_scans, strict=False
    ):
        assert rotation_scan.omega_start_deg == expected["omega_start_deg"]
        assert rotation_scan.chi_start_deg == expected["chi_start_deg"]
        assert rotation_scan.x_start_um == expected["x_start_um"]
        assert rotation_scan.y_start_um == expected["y_start_um"]
        assert rotation_scan.z_start_um == expected["z_start_um"]
        assert rotation_scan.nexus_vds_start_img == expected["nexus_vds_start_img"]


@patch("mx_bluesky.common.parameters.components.os.makedirs")
def test_load_centre_collect_creates_storage_directory_if_not_present(
    mock_makedirs,
):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json"
    )
    LoadCentreCollect(**params)

    mock_makedirs.assert_has_calls(
        [
            call(
                "/tmp/dls/i03/data/2024/cm31105-4/auto/123458/xraycentring",
                exist_ok=True,
            )
        ],
        any_order=True,
    )
    mock_makedirs.assert_has_calls(
        [call("/tmp/dls/i03/data/2024/cm31105-4/auto/123458/", exist_ok=True)],
        any_order=True,
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan"
)
def test_box_size_passed_through_to_gridscan(
    mock_detect_grid: MagicMock,
    composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
):
    load_centre_collect_params.robot_load_then_centre.box_size_um = 25

    RE(
        load_centre_collect_full(
            composite, load_centre_collect_params, oav_parameters_for_rotation
        )
    )
    detect_grid_call = mock_detect_grid.mock_calls[0]
    assert detect_grid_call.args[1].box_size_um == 25


@patch(
    "mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan.multi_rotation_scan",
    return_value=iter([]),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan.robot_load_then_xray_centre",
    return_value=iter([]),
)
def test_load_centre_collect_full_collects_at_current_location_if_no_xray_centring_required(
    _: MagicMock,
    mock_rotation_scan: MagicMock,
    composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_read_handler_for(composite.smargon.x, 1.1)
    sim_run_engine.add_read_handler_for(composite.smargon.y, 2.2)
    sim_run_engine.add_read_handler_for(composite.smargon.z, 3.3)

    sim_run_engine.simulate_plan(
        load_centre_collect_full(
            composite, load_centre_collect_params, oav_parameters_for_rotation
        )
    )

    rotation_scans = mock_rotation_scan.call_args.args[1].rotation_scans
    assert len(rotation_scans) == 1
    assert rotation_scans[0].x_start_um == 1.1
    assert rotation_scans[0].y_start_um == 2.2
    assert rotation_scans[0].z_start_um == 3.3



from __future__ import annotations

import os
from abc import abstractmethod
from collections.abc import Sequence
from enum import StrEnum
from pathlib import Path
from typing import Literal, SupportsInt, cast

from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.detector import (
    DetectorParams,
    TriggerMode,
)
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
    model_validator,
)
from pydantic_extra_types.semantic_version import SemanticVersion
from scanspec.core import AxesPoints
from semver import Version

from mx_bluesky.common.parameters.constants import (
    TEST_MODE,
    DetectorParamConstants,
    GridscanParamConstants,
)

PARAMETER_VERSION = Version.parse("5.3.0")


class RotationAxis(StrEnum):
    OMEGA = "omega"
    PHI = "phi"
    CHI = "chi"
    KAPPA = "kappa"


class XyzAxis(StrEnum):
    X = "sam_x"
    Y = "sam_y"
    Z = "sam_z"


class IspybExperimentType(StrEnum):
    # Enum values from ispyb column data type
    SAD = "SAD"  # at or slightly above the peak
    SAD_INVERSE_BEAM = "SAD - Inverse Beam"
    OSC = "OSC"  # "native" (in the absence of a heavy atom)
    COLLECT_MULTIWEDGE = (
        "Collect - Multiwedge"  # "poorly determined" ~ EDNA complex strategy???
    )
    MAD = "MAD"
    HELICAL = "Helical"
    MULTI_POSITIONAL = "Multi-positional"
    MESH = "Mesh"
    BURN = "Burn"
    MAD_INVERSE_BEAM = "MAD - Inverse Beam"
    CHARACTERIZATION = "Characterization"
    DEHYDRATION = "Dehydration"
    TOMO = "tomo"
    EXPERIMENT = "experiment"
    EM = "EM"
    PDF = "PDF"
    PDF_BRAGG = "PDF+Bragg"
    BRAGG = "Bragg"
    SINGLE_PARTICLE = "single particle"
    SERIAL_FIXED = "Serial Fixed"
    SERIAL_JET = "Serial Jet"
    STANDARD = "Standard"  # Routine structure determination experiment
    TIME_RESOLVED = "Time Resolved"  # Investigate the change of a system over time
    DLS_ANVIL_HP = "Diamond Anvil High Pressure"  # HP sample environment pressure cell
    CUSTOM = "Custom"  # Special or non-standard data collection
    XRF_MAP = "XRF map"
    ENERGY_SCAN = "Energy scan"
    XRF_SPECTRUM = "XRF spectrum"
    XRF_MAP_XAS = "XRF map xas"
    MESH_3D = "Mesh3D"
    SCREENING = "Screening"
    STILL = "Still"
    SSX_CHIP = "SSX-Chip"
    SSX_JET = "SSX-Jet"
    METAL_ID = "Metal ID"

    # Aliases for historic hyperion experiment type mapping
    ROTATION = "SAD"
    GRIDSCAN_2D = "mesh"
    GRIDSCAN_3D = "Mesh3D"


class MxBlueskyParameters(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )

    def __hash__(self) -> int:
        return self.model_dump_json().__hash__()

    parameter_model_version: SemanticVersion

    @field_validator("parameter_model_version")
    @classmethod
    def _validate_version(cls, version: SemanticVersion):
        assert version >= SemanticVersion(major=PARAMETER_VERSION.major), (
            f"Parameter version too old! This version of hyperion uses {PARAMETER_VERSION}"
        )
        assert version <= SemanticVersion(major=PARAMETER_VERSION.major + 1), (
            f"Parameter version too new! This version of hyperion uses {PARAMETER_VERSION}"
        )
        return version


class WithSnapshot(BaseModel):
    snapshot_directory: Path
    snapshot_omegas_deg: list[float] | None = None

    @property
    def take_snapshots(self) -> bool:
        return bool(self.snapshot_omegas_deg)


class WithOptionalEnergyChange(BaseModel):
    demand_energy_ev: float | None = Field(default=None, gt=0)


class WithVisit(BaseModel):
    beamline: str = Field(default="BL03I", pattern=r"BL\d{2}[BIJS]")
    visit: str = Field(min_length=1)
    det_dist_to_beam_converter_path: str = Field(
        default=DetectorParamConstants.BEAM_XY_LUT_PATH
    )
    insertion_prefix: str = "SR03S" if TEST_MODE else "SR03I"
    detector_distance_mm: float | None = Field(default=None, gt=0)


class DiffractionExperiment(
    MxBlueskyParameters, WithSnapshot, WithOptionalEnergyChange, WithVisit
):
    """For all experiments which use beam"""

    file_name: str
    exposure_time_s: float = Field(gt=0)
    comment: str = Field(default="")
    trigger_mode: TriggerMode = Field(default=TriggerMode.FREE_RUN)
    run_number: int | None = Field(default=None, ge=0)
    selected_aperture: ApertureValue | None = Field(default=None)
    transmission_frac: float = Field(default=0.1)
    ispyb_experiment_type: IspybExperimentType
    storage_directory: str
    use_roi_mode: bool = Field(default=GridscanParamConstants.USE_ROI)

    @model_validator(mode="before")
    @classmethod
    def validate_directories(cls, values):
        os.makedirs(values["storage_directory"], exist_ok=True)

        values["snapshot_directory"] = values.get(
            "snapshot_directory",
            Path(values["storage_directory"], "snapshots").as_posix(),
        )
        return values

    @property
    def num_images(self) -> int:
        return 0

    @property
    @abstractmethod
    def detector_params(self) -> DetectorParams: ...


class WithScan(BaseModel):
    """For experiments where the scan is known"""

    @property
    @abstractmethod
    def scan_points(self) -> AxesPoints: ...

    @property
    @abstractmethod
    def num_images(self) -> int: ...


class WithPandaGridScan(BaseModel):
    """For experiments which use a PandA for constant-motion grid scans"""

    panda_runup_distance_mm: float = Field(
        default=GridscanParamConstants.PANDA_RUN_UP_DISTANCE_MM
    )


class SplitScan(BaseModel):
    @property
    @abstractmethod
    def scan_indices(self) -> Sequence[SupportsInt]:
        """Should return the first index of each scan (i.e. for each nexus file)"""
        ...


class WithSample(BaseModel):
    sample_id: int
    sample_puck: int | None = None
    sample_pin: int | None = None


class DiffractionExperimentWithSample(DiffractionExperiment, WithSample): ...


class MultiXtalSelection(BaseModel):
    name: str


class TopNByMaxCountSelection(MultiXtalSelection):
    name: Literal["TopNByMaxCount"] = "TopNByMaxCount"  #  pyright: ignore [reportIncompatibleVariableOverride]
    n: int


class WithCentreSelection(BaseModel):
    select_centres: TopNByMaxCountSelection = Field(
        discriminator="name", default=TopNByMaxCountSelection(n=1)
    )

    @property
    def selection_params(self) -> MultiXtalSelection:
        """A helper property because pydantic does not allow polymorphism with base classes
        # only type unions"""
        cast1 = cast(MultiXtalSelection, self.select_centres)
        return cast1


class OptionalXyzStarts(BaseModel):
    x_start_um: float | None = None
    y_start_um: float | None = None
    z_start_um: float | None = None


class XyzStarts(BaseModel):
    x_start_um: float
    y_start_um: float
    z_start_um: float

    def _start_for_axis(self, axis: XyzAxis) -> float:
        match axis:
            case XyzAxis.X:
                return self.x_start_um
            case XyzAxis.Y:
                return self.y_start_um
            case XyzAxis.Z:
                return self.z_start_um


class OptionalGonioAngleStarts(BaseModel):
    omega_start_deg: float | None = None
    phi_start_deg: float | None = None
    chi_start_deg: float | None = None
    kappa_start_deg: float | None = None



from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger, Pilatus


def test_eiger():
    eig = Eiger()
    assert eig.image_size_mm == (233.1, 244.65)


def test_pilatus():
    pil = Pilatus()
    assert pil.image_size_mm == (423.636, 434.644)



from unittest.mock import ANY, MagicMock, call, patch

import bluesky.plan_stubs as bps
import cv2 as cv
import pytest
from dodal.devices.i24.pmac import PMAC
from dodal.devices.oav.oav_detector import OAV
from ophyd_async.testing import get_mock_put

from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick import (
    _calculate_zoom_calibrator,
    _get_beam_centre,
    _move_on_mouse_click_plan,
    onMouse,
    update_ui,
)

ZOOMCALIBRATOR = 6


def fake_generator(value):
    yield from bps.null()
    return value


@pytest.mark.parametrize(
    "beam_position, expected_xmove, expected_ymove",
    [
        (
            (15, 10),
            "#1J:-" + str(10 * 15 * ZOOMCALIBRATOR),
            "#2J:" + str(10 * 10 * ZOOMCALIBRATOR),
        ),
        (
            (475, 309),
            "#1J:-" + str(10 * 475 * ZOOMCALIBRATOR),
            "#2J:" + str(10 * 309 * ZOOMCALIBRATOR),
        ),
        (
            (638, 392),
            "#1J:-" + str(10 * 638 * ZOOMCALIBRATOR),
            "#2J:" + str(10 * 392 * ZOOMCALIBRATOR),
        ),
    ],
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick._get_beam_centre"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick._calculate_zoom_calibrator"
)
def test_move_on_mouse_click_gets_beam_position_and_sends_correct_str(
    fake_zoom_calibrator: MagicMock,
    fake_get_beam_pos: MagicMock,
    beam_position: tuple,
    expected_xmove: str,
    expected_ymove: str,
    pmac: PMAC,
    RE,
):
    fake_zoom_calibrator.side_effect = [fake_generator(ZOOMCALIBRATOR)]
    fake_get_beam_pos.side_effect = [fake_generator(beam_position)]
    fake_oav: OAV = MagicMock(spec=OAV)
    RE(_move_on_mouse_click_plan(fake_oav, pmac, (0, 0)))

    mock_pmac_str = get_mock_put(pmac.pmac_string)
    mock_pmac_str.assert_has_calls(
        [
            call(expected_xmove, wait=True),
            call(expected_ymove, wait=True),
        ]
    )


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick._move_on_mouse_click_plan"
)
def test_onMouse_runs_plan_on_click(fake_move_plan: MagicMock, pmac: PMAC, RE):
    fake_oav: OAV = MagicMock(spec=OAV)
    onMouse(cv.EVENT_LBUTTONUP, 0, 0, "", param=[RE, pmac, fake_oav])
    fake_move_plan.assert_called_once()


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick.bps.rd")
def test_get_beam_centre(fake_read: MagicMock, RE):
    fake_read.side_effect = [fake_generator(10), fake_generator(2)]
    fake_oav: OAV = MagicMock(spec=OAV)
    fake_oav.beam_centre_i = fake_oav.beam_centre_j = MagicMock()
    res = RE(_get_beam_centre(fake_oav)).plan_result

    assert res == (10, 2)


@pytest.mark.parametrize(
    "zoom_percentage, expected_calibrator", [(1, 1.256), (20, 0.805), (50, 0.375)]
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick.bps.rd")
def test_calculate_zoom_calibrator(
    fake_read: MagicMock, zoom_percentage: int, expected_calibrator: float, RE
):
    fake_read.side_effect = [fake_generator(zoom_percentage)]
    fake_oav: OAV = MagicMock(spec=OAV)
    fake_oav.zoom_controller = MagicMock()
    res = RE(_calculate_zoom_calibrator(fake_oav)).plan_result  # type: ignore

    assert res == pytest.approx(expected_calibrator, abs=1e-3)


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick.cv")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick._get_beam_centre"
)
def test_update_ui_uses_correct_beam_centre_for_ellipse(fake_beam_pos, fake_cv, RE):
    mock_frame = MagicMock()
    mock_oav = MagicMock()
    fake_beam_pos.side_effect = [fake_generator([15, 10])]
    update_ui(mock_oav, mock_frame, RE)
    fake_cv.ellipse.assert_called_once()
    fake_cv.ellipse.assert_has_calls(
        [call(ANY, (15, 10), (12, 8), 0.0, 0.0, 360, (0, 255, 255), thickness=2)]
    )
